import textwrap

from conan.test.utils.tools import TestClient


class TestConanfileRepeatedGenerators:

    def test_conanfile_txt(self):
        conanfile = textwrap.dedent("""
            [generators]
            CMakeDeps
            CMakeDeps
        """)

        t = TestClient()
        t.save({'conanfile.txt': conanfile})
        t.run("install conanfile.txt")
        assert str(t.out).count("Generator 'CMakeDeps' calling 'generate()'") == 1

    def test_conanfile_py(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Recipe(ConanFile):
                settings = "build_type"
                generators = "CMakeDeps", "CMakeDeps"
        """)
        t = TestClient()
        t.save({'conanfile.py': conanfile})
        t.run("install conanfile.py")
        assert str(t.out).count("Generator 'CMakeDeps' calling 'generate()'") == 1

    def test_python_requires_inheritance(self):
        pyreq = textwrap.dedent("""
            from conan import ConanFile

            class Recipe(ConanFile):
                pass

            class BaseConan(object):
                generators = "CMakeDeps",
        """)
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Recipe(ConanFile):
                settings = "build_type"
                python_requires = "base/1.0"
                python_requires_extend = "base.BaseConan"
                settings = "build_type"
                generators = "CMakeDeps",

                def init(self):
                    base = self.python_requires["base"].module.BaseConan
                    self.generators = base.generators + self.generators
        """)

        t = TestClient()
        t.save({'pyreq.py': pyreq, 'conanfile.py': conanfile})
        t.run("export pyreq.py --name=base --version=1.0")
        t.run("install conanfile.py")
        assert str(t.out).count("Generator 'CMakeDeps' calling 'generate()'") == 1

    def test_duplicated_generator_in_member_and_attribue(self):
        """
        Ensure we raise an error when a generator is present both in the generators attribute
        and instanced in the generate() method by the user, which we didn't use to do before 2.0
        """
        conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain

        class Recipe(ConanFile):
            generators = "CMakeToolchain"

            def generate(self):
                tc = CMakeToolchain(self)
                tc.generate()
        """)

        t = TestClient()
        t.save({'conanfile.py': conanfile})
        # This used to not throw any errors
        t.run("install .", assert_error=True)
        assert "was instantiated in the generate() method too" in t.out



import os
import shutil
import textwrap

from conan.test.utils.tools import TestClient


class TestEditableImport:

    def test_copy_from_dep_in_generate(self):
        """ as we are removing the imports explicit functionality, test if the editable
        can still work for the "imports" case. It seem possible if:

        - "dep" package, both in cache and editable mode, defines correctly its layout
        - Tricky: "dep" package musth use self.source_folder+res instead of layout to package()
        - IMPORTANT: Consumers should use dep.cpp_info.resdirs[0], not dep.package_folder
        """
        t = TestClient()
        dep = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import copy
            class Pkg(ConanFile):
                exports_sources = "*"
                def layout(self):
                    self.folders.source = "src"
                    self.cpp.source.resdirs = ["res"]
                def package(self):
                    resdir = os.path.join(self.source_folder, "res")
                    copy(self, "*", resdir, os.path.join(self.package_folder, "data"))
                def package_info(self):
                    self.cpp_info.resdirs = ["data"]
            """)
        consumer = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import copy
            class Pkg(ConanFile):
                requires = "dep/0.1"

                def generate(self):
                    dep = self.dependencies["dep"]
                    resdir = dep.cpp_info.resdirs[0]
                    copy(self, "*", resdir, os.path.join(self.build_folder, "imports"))
            """)
        t.save({'dep/conanfile.py': dep,
                'dep/src/res/myfile.txt': "mydata",
                "consumer/conanfile.py": consumer})

        t.run("create dep --name=dep --version=0.1")
        t.run("install consumer")
        assert t.load("consumer/imports/myfile.txt") == "mydata"

        t.run("remove * -c")
        t.run('editable add dep --name=dep --version=0.1')
        shutil.rmtree(os.path.join(t.current_folder, "consumer", "imports"))
        t.run("install consumer")
        assert t.load("consumer/imports/myfile.txt") == "mydata"



diff_html = r"""
{% macro render_sidebar_folder(folder, folder_info) %}
    {%- for name, sub_folder_info in folder_info["folders"].items() %}
        {% set folder_name = folder + "/" + name %}
        <li>
            <details open class="folder">
                <summary>{{ name }}</summary>
                <ul>
                    {{ render_sidebar_folder(folder_name, sub_folder_info) }}
                </ul>
            </details>
        </li>
    {%- endfor %}
    {%- for name, file_info in folder_info["files"].items() %}
        {% set file_type = "renamed" if file_info["renamed_to"] else (
                           "deleted" if file_info["is_deleted"] else (
                           "new" if file_info["is_new"] else "old")) %}
        <li class="file file-{{ file_type }}"
            data-path="{{ file_info["relative_path"] }}"
            data-type="{{ file_type }}">
            <a href="#diff_{{- safe_filename(file_info["filename"]) -}}"
                onclick="setDataIsLinked(event)" draggable="false"
                class="side-link"
                title="{{ replace_cache_paths(file_info["relative_path"]) | replace("(old)/", "") | replace("(new)/", "") }}">
                {% if file_info["renamed_to"] %}
                    {{ file_info["renamed_to"].split("/")[1:][-1] }}
                {% else %}
                    {{ name }}
                {% endif %}
            </a>
        </li>
    {%- endfor %}
{% endmacro %}

{% macro render_diff_folder(folder_info) %}
    {%- for name, sub_folder_info in folder_info["folders"].items() %}
        {{ render_diff_folder(sub_folder_info) }}
    {%- endfor %}
    {%- for name, file_info in folder_info["files"].items() %}
        {% set filename = file_info["filename"] %}

        <div id="diff_{{ safe_filename(filename) }}" data-path="{{ filename }}" class="diff-container">
            <div class="diff-content">
                <details open class="diff-details">
                    <summary class="diff-summary">
                        <b id="diff_{{ safe_filename(filename) }}_filename" class="filename" data-replaced-paths="">
                            <span>{{ replace_cache_paths(filename) | replace("(old)/", "") | replace("(new)/", "") }}</span>
                            {% if file_info["renamed_to"] %}
                                &nbsp;&#x2192&nbsp;
                                <span>{{ replace_cache_paths(file_info["renamed_to"]) | replace("(old)/", "") | replace("(new)/", "") }}</span>
                            {% endif %}
                        </b>
                        <div class="changes-count-container"></div>
                    </summary>
                    <div class="diff-lines">
                    </div>
                </details>
            </div>
        </div>
    {%- endfor %}
{% endmacro %}
<html lang="en">
    <head>
        <meta charset="utf-8">
        <title>Diff report for {{ old_reference }} - {{ new_reference }}</title>
        <style>
            /* --- Colors --- */
            :root {
                --body-bgColor: #f8f8f8;
                --sidebar-bgColor: #f4f4f466;
                --sidebar-borderColor: #ccc;
                --sidebar-contents-bgColor: #f4f4f4;
                --content-bgColor: #f8f8f8;
                --search-area-borderColor: #ccc;
                --search-field-borderColor: #ccc;
                --file-list-borderColor: #ddd;
                --folder-summary-hover-bgColor: #e0e0e033;
                --folder-ul-hover-borderColor: #00000066;
                --sidebar-li-a-hover-bgColor: #e0e0e0;
                --sidebar-button-hover-bgColor: var(--sidebar-li-a-hover-bgColor);
                --sidebar-link-color: black;
                --sidebar-link-hover-color: var(--sidebar-link-color);
                --sidebar-link-visited-color: var(--sidebar-link-color);
                --sidebar-file-new-color: green;
                --sidebar-file-old-color: gray;
                --sidebar-file-deleted-color: red;
                --diff-content-borderColor: black;
                --diff-content-bgColor: white;
                --diff-container-linked-borderColor: #0078d7;
                --diff-summary-borderColor: #ccc;
                --diff-summary-bgColor: #f8f8f8;
                --diff-summary-hover-bgColor: #f0f0f0;
                --new-lines-count-color: green;
                --old-lines-count-color: black;
                --context-line-color: #888;
                --context-chunk-header-bgColor: #cef8ff;
                --context-chunk-header-color: var(--context-line-color);
                --added-line-bgColor: #cbfcd9;
                --added-line-color: black;
                --deleted-line-bgColor: #ffebe9;
                --deleted-line-color: black;
                --line-number-added-bgColor: #76ffbb;
                --line-number-deleted-bgColor: #fdb9c1;
            }

            /* --- Global Styles --- */

            body {
                font-family: monospace;
                margin: 0px;
                background-color: var(--body-bgColor);
            }

            /* --- Main Layout --- */

            .container {
                display: flex;
                height: 100%;
                overflow: scroll;
            }

            .sidebar {
                width: 17%;
                min-width: 10%;
                max-width: 33%;
                padding: 10px;
                overflow: scroll;
                background: var(--sidebar-bgColor);
                border-right: 1px solid var(--sidebar-borderColor);
                resize: horizontal;
                position: sticky;
                top: 0;
            }

            .content {
                padding: 20px;
                background: var(--content-bgColor);
                width: 100%;
            }

            /* --- Sidebar & File Tree --- */

            #sidebar-contents {
                background: var(--sidebar-contents-bgColor);
                border-radius: 7px;
                overflow-y: hidden;
                padding-top: 5px;
            }

            .sidebar-reveal {
                display: none;
                position: sticky;
                top: 10px;
            }

            .search-area {
                border-bottom: 1px solid var(--search-area-borderColor);
            }

            .search-header {
                display: flex;
                justify-content: space-between;
            }

            .search-field {
                border: 1px solid var(--search-field-borderColor);
                border-radius: 5px;
                padding: 5px;
                margin: 5px;
                width: 80%;
            }

            .file-tree-controls {
                border-bottom: 1px solid var(--search-area-borderColor);
                padding: 5px;
                display: flex;
                justify-content: space-between;
                align-items: center;
            }

            .file-tree-controls .folder-collapse button {
                display: inline-block;
                line-height: 0.7;
            }

            .file-tree-controls button,
            .sidebar-reveal button,
            .search-header button {
                cursor: pointer;
                border: 0px solid var(--search-field-borderColor);
                border-radius: 5px;
                background: none;
                padding: 5px;
                min-width: 3ch;
            }

            .file-tree-controls button:hover,
            .sidebar-reveal button:hover,
            .search-header button:hover {
                background-color: var(--sidebar-li-a-hover-bgColor);
            }

            .file-tree-more {
                display: none;
                padding: 5px;
                border-bottom: 1px solid var(--search-area-borderColor);
            }

            .file-tree-more-option {
                display: block;
            }

            .file-list {
                padding-left: 10px;
                width: 100%;
                overflow-x: clip;
            }

            .file-list ul li {
                width: 100%;
            }

            .file-list li ul {
                border-left: 1px solid var(--file-list-borderColor);
                margin-left: 3px;
            }

            li ul {
                padding-left: 1ch;
            }

            details.folder {
                text-wrap: nowrap;
            }

            .folder > summary {
                cursor: pointer;
                list-style: none;
            }

            .folder > summary:hover {
                background-color: var(--folder-summary-hover-bgColor);
            }

            .folder:not(:open) > summary:before {
                content: "\1F4C1";
                display: inline-block;
                margin-right: 3px;
            }

            .folder:open > summary:before {
                content: "\1F4C2";
                display: inline-block;
                margin-right: 3px;
            }

            details.folder ul:hover {
                border-left: 1px solid var(--folder-ul-hover-borderColor);
            }

            .sidebar li {
                line-height: 1.8;
                list-style: none;
                list-style-position: inside;
                user-select: none;
            }

            .sidebar li a {
                text-decoration: none;
                padding: 5px;
                color: var(--sidebar-link-color);
            }

            .sidebar li a:hover {
                text-decoration: none;
                border-radius: 5px;
                background-color: var(--sidebar-li-a-hover-bgColor);
                padding: 5px;
                color: var(--sidebar-link-hover-color);
            }

            .sidebar li a:visited {
                color: var(--sidebar-link-visited-color);
            }

            .side-link {
                text-wrap: nowrap;
            }

            /* File Status Indicators */
            .sidebar li.file-new,
            .sidebar li.file-old,
            .sidebar li.file-deleted,
            .sidebar li.file-renamed {
                list-style: none;
                padding-left: 0;
            }

            .sidebar li.file-new:before {
                content: "+";
                color: var(--sidebar-file-new-color);
                font-weight: bold;
            }

            .sidebar li.file-old:before {
                content: "\00B1";
                color: var(--sidebar-file-old-color);
            }

            .sidebar li.file-deleted:before {
                content: "-";
                color: var(--sidebar-file-deleted-color);
                font-weight: bold;
            }

            .sidebar li.file-renamed:before {
                content: "\2192";
                color: var(--sidebar-file-old-color);
                font-weight: bold;
            }

            /* --- Diff View Components --- */

            .diff-container {
                scroll-margin-top: 10px;
            }

            .diff-content {
                padding-bottom: 7px;
                border: 1px solid var(--diff-content-borderColor);
                border-radius: 7px;
                margin-bottom: 10px;
                background-color: var(--diff-content-bgColor);
            }

            .diff-container[data-is-linked="true"] .diff-content {
                border: 2px solid var(--diff-container-linked-borderColor);
            }

            details.diff-details summary.diff-summary {
                cursor: pointer;
                display: flex;
                justify-content: space-between;
                align-items: center;
                border-bottom: 1px solid var(--diff-summary-borderColor);
                padding: 5px 0px;
                position: sticky;
                top: 0;
                background-color: var(--diff-summary-bgColor);
                border-radius: 7px 7px 0px 0px;
            }

            details.diff-details summary.diff-summary:hover {
                background-color: var(--diff-summary-hover-bgColor);
            }

            details:open .diff-summary .filename:before {
                content: "\25BC";
                display: inline-block;
            }

            details:not(:open) .diff-summary .filename:before {
                content: "\25B6";
                display: inline-block;
            }

            .diff-header {
                padding: 0px 5px 5px 5px;
            }

            .filename {
                font-size: 1.2em;
                padding-left: 10px;
            }

            .changes-count-container {
                font-size: 0.9em;
                padding-right: 10px;
            }

            .new-lines-count {
                color: var(--new-lines-count-color);
                font-weight: bold;
            }

            .old-lines-count {
                color: var(--old-lines-count-color);
                font-weight: bold;
            }

            /* --- Diff Line Styles --- */

            .content span {
                white-space: pre-wrap;
            }

            .context-chunk-header {
                list-style: none;
                background-color: var(--context-chunk-header-bgColor);
                color: var(--context-chunk-header-color);
                line-height: 2;
                cursor: pointer;
            }

            details:open .context-chunk-header .line-number:before {
                content: "\25BC";
                display: inline-block;
            }

            details:not(:open) .context-chunk-header .line-number:before {
                content: "\25B6";
                display: inline-block;
            }

            .diff-lines {
                line-break: anywhere;
            }

            .line-number {
                width: 4ch;
                min-width: 4ch;
                display: inline-block;
                text-align: center;
                user-select: none;
            }

            .context-line {
                color: var(--context-line-color);
            }

            .add {
                background-color: var(--added-line-bgColor);
                color: var(--added-line-color);
            }

            .del {
                background-color: var(--deleted-line-bgColor);
                color: var(--deleted-line-color);
            }

            .add,
            .del,
            .context-line {
                height: 100%;
            }

            .diff-line {
                display: flex;
                box-sizing: border-box;
                line-height: 1.5em;
            }

            .line-number.add {
                background-color: var(--line-number-added-bgColor);
            }

            .line-number.del {
                background-color: var(--line-number-deleted-bgColor);
            }

            .line-number.add,
            .line-number.del {
                height: auto;
            }

            .diff-symbol {
                display: inline-block;
                width: 1ch;
                user-select: none;
            }

            /* --- Utility & Page States --- */

            #empty_result {
                justify-content: center;
                align-items: center;
                color: black;
                font-weight: bold;
                font-size: 4em;
                text-align: center;
            }
        </style>
        <script>

            const data = {{ content | tojson | safe }};

            const oldPattern = "{{ src_prefix[:-1] }}{{ old_cache_path }}";
            const newPattern = "{{ dst_prefix[:-1] }}{{ new_cache_path }}";

            function extractLineNumbers(hunkHeader) {
                const regex = /@@ -(\d+),\d+ \+(\d+),\d+ @@/;
                const match = hunkHeader.match(regex);
                if (!match) {
                    return [0, 0];
                }
                return [parseInt(match[1]), parseInt(match[2])];
            }


            function makeDiffLines(lines) {
                const element = document.createElement("div");
                let seen_header = false;
                let new_line_index = 0;
                let old_line_index = 0;
                let new_line_count = 0;
                let old_line_count = 0;
                const headerDiv = document.createElement("div");
                let currentDetails = null;
                for (let i = 0; i < lines.length; i++) {
                    const line = lines[i];
                    let spanLine = document.createElement("span");
                    const lineDiv = document.createElement("div");
                    lineDiv.className = "diff-line";
                    let shouldAddLine = true;
                    if (line.startsWith("+++")) {
                        seen_header = true;
                        spanLine.className = "add";
                        spanLine.textContent = line.replace(newPattern, "(new)");
                        headerDiv.appendChild(spanLine);
                        continue;
                    } else if (line.startsWith("---")) {
                        spanLine.className = "del";
                        spanLine.textContent = line.replace(oldPattern, "(old)");
                        headerDiv.appendChild(spanLine);
                        continue;
                    } else if (line.startsWith("@@")) {
                        currentDetails = document.createElement("details");
                        currentDetails.open = true;

                        const summary = document.createElement("summary");
                        summary.className = "context-chunk-header";
                        const summaryArrow = document.createElement("span");
                        summaryArrow.className = "line-number";
                        const summaryText = document.createElement("span");
                        summaryText.textContent = line;

                        summary.appendChild(summaryArrow);
                        summary.appendChild(summaryText);

                        currentDetails.appendChild(summary);
                        element.appendChild(currentDetails);
                        shouldAddLine = false;

                        const lineNumbers = extractLineNumbers(line);
                        old_line_index = lineNumbers[0];
                        new_line_index = lineNumbers[1];
                    } else if (line.startsWith("+")) {
                        const spanSymbol = document.createElement("span");
                        spanSymbol.textContent = "+";
                        spanSymbol.className = "diff-symbol";
                        spanLine.className = "add";
                        spanLine.textContent = line.substring(1);
                        spanLine.prepend(spanSymbol);

                        const lineNumberSpan = document.createElement("span");
                        lineNumberSpan.className = "line-number add";
                        lineNumberSpan.textContent = new_line_index;
                        lineDiv.appendChild(lineNumberSpan);

                        new_line_index += 1;
                        new_line_count += 1;
                    } else if (line.startsWith("-")) {
                        const spanSymbol = document.createElement("span");
                        spanSymbol.textContent = "-";
                        spanSymbol.className = "diff-symbol";
                        spanLine.className = "del";
                        spanLine.textContent = line.substring(1);
                        spanLine.prepend(spanSymbol);

                        const lineNumberSpan = document.createElement("span");
                        lineNumberSpan.className = "line-number del";
                        lineNumberSpan.textContent = old_line_index;
                        lineDiv.appendChild(lineNumberSpan);

                        old_line_index += 1;
                        old_line_count += 1;
                    } else {
                        spanLine.className = "context-line";
                        if (!seen_header) {
                            spanLine.textContent = line.replace(oldPattern, "(old)").replace(newPattern, "(new)");
                            headerDiv.appendChild(spanLine);
                            headerDiv.appendChild(document.createElement("br"));
                            continue;
                        } else {
                            const spanSymbol = document.createElement("span");
                            spanSymbol.className = "diff-symbol";
                            spanLine.textContent = line;
                            spanLine.prepend(spanSymbol);
                        }

                        const lineNumberSpan = document.createElement("span");
                        lineNumberSpan.className = "line-number context-line";
                        lineNumberSpan.textContent = new_line_index;
                        lineDiv.appendChild(lineNumberSpan);

                        new_line_index += 1;
                        old_line_index += 1;
                    }
                    if (shouldAddLine) {
                        lineDiv.appendChild(spanLine);

                        currentDetails.appendChild(lineDiv);
                        //currentDetails.appendChild(document.createElement("br"));
                    }
                }
                if (!seen_header) {
                    element.appendChild(headerDiv);
                }
                return [element, new_line_count, old_line_count];
            }

            function createChangesCountElement(new_count, old_count) {
                const changes = document.createElement("span");
                changes.className = "changes-count";
                changes.innerHTML = `<span class="new-lines-count">+${new_count}</span> <span class="old-lines-count">-${old_count}</span>`;
                return changes;
            }


            function intersectionCallback(entries) {
              entries.forEach((entry) => {
                if (entry.isIntersecting) {
                    let elem = entry.target;
                    const path = elem.dataset.path;
                    const [lines, new_count, old_count] = makeDiffLines(data[path]);
                    const diffLines = elem.querySelector(".diff-lines")

                    // If we're scrolling up, new lines are added to the top, so we need to
                    // preserve the scroll position relative to the bottom of the new content
                    const prevRect = elem.getBoundingClientRect();


                    diffLines.appendChild(lines);

                    if (new_count !== 0 || old_count !== 0) {
                        elem.querySelector(".changes-count-container").appendChild(createChangesCountElement(new_count, old_count));
                    }

                    if (elem.getAttribute("data-is-linked") === "true") {
                        // We need to scroll to the element again now that its height has changed
                        elem.scrollIntoView({block: "start", inline: "nearest", behavior: "instant"});
                    } else {
                        if (prevRect.top < 0) {
                            const prevBottom = prevRect.bottom;
                            const newBottom = elem.getBoundingClientRect().bottom;
                            const container = document.querySelector('.container');
                            container.scroll(0, container.scrollTop + (newBottom - prevBottom));
                        }
                    }

                    observer.unobserve(elem);
                }
              });
            }

            const options = {
                root: document.querySelector('.content'),
                rootMargin: "0px",
                scrollMargin: "0px",
                threshold: 0.05,
            };

            const observer = new IntersectionObserver(intersectionCallback, options);

            document.addEventListener("DOMContentLoaded", (e) => {
                setDataIsLinked(null);
                document.querySelectorAll('.diff-container').forEach((section) => {
                    observer.observe(section);
                });
            });

            function debounce(func, delay) {
                let timeout;
                return function(...args) {
                    const context = this;
                    clearTimeout(timeout);
                    timeout = setTimeout(() => {
                        func.apply(context, args);
                    }, delay);
                };
            }
            let includeSearchQuery = "";
            let excludeSearchQuery = "";

            async function onSearchInput(event) {
                const sidebar = document.querySelectorAll(".sidebar li");
                const fileList = document.querySelector(".file-list");
                const content = document.querySelectorAll(".content .diff-container .diff-content");
                const searchingIcon = document.getElementById("searching_icon");

                searchingIcon.style.display = "inline-block";

                let emptySearch = true;
                let includedFiles = 0;

                const typeVisibility = {
                    "renamed": document.getElementById("show-moved-files").checked,
                    "deleted": document.getElementById("show-deleted-files").checked,
                    "new": document.getElementById("show-new-files").checked,
                    "old": document.getElementById("show-old-files").checked,
                };

                sidebar.forEach(async function(item) {
                    if (item.dataset.path === undefined) {
                        // A folder, those are handled later
                        return;
                    }
                    const text = item.dataset.path.toLowerCase();
                    const shouldInclude = includeSearchQuery === "" || text.includes(includeSearchQuery);
                    let shouldExclude = excludeSearchQuery !== "" && text.includes(excludeSearchQuery);
                    const associatedId = item.querySelector("a").getAttribute("href").substring(1)
                    const contentItem = document.getElementById(associatedId);

                    const fileType = item.dataset.type;
                    const isTypeVisible = typeVisibility[fileType] !== false;

                    shouldExclude = shouldExclude || !isTypeVisible;

                    if (shouldInclude) {
                        if (shouldExclude) {
                            item.style.display = "none";
                            contentItem.style.display = "none";
                        } else {
                            includedFiles += 1;
                            item.style.display = "list-item";
                            contentItem.style.display = "block";
                            emptySearch = false;
                        }
                    } else {
                        item.style.display = "none";
                        contentItem.style.display = "none";
                    }

                });

                searchingIcon.style.display = "none";
                const emptySearchTag = document.getElementById("empty_search");
                const emptyResultTag = document.getElementById("empty_result");
                if (emptySearch) {
                    emptySearchTag.style.display = "block";
                    emptyResultTag.style.display = "block";
                    fileList.style.display = "none";
                } else {
                    emptySearchTag.style.display = "none";
                    emptyResultTag.style.display = "none";
                    fileList.style.display = "block";
                }

                const fileCountTag = document.getElementById("file-count");
                fileCountTag.textContent = includedFiles;

                const allDetails = document.querySelectorAll(".sidebar details.folder");
                allDetails.forEach(function(details) {
                    details.style.display = "none";
                    details.querySelectorAll("li.file").forEach(function(li) {
                        if (li.style.display !== "none") {
                            details.style.display = "block";
                            return;
                        }
                    });
                });

            }

            const debouncedOnSearchInput = debounce(onSearchInput, 300);

            async function onExcludeSearchInput(event) {
                excludeSearchQuery = event.currentTarget.value.toLowerCase();
                debouncedOnSearchInput(event);
            }

            async function onIncludeSearchInput(event) {
                includeSearchQuery = event.currentTarget.value.toLowerCase();
                debouncedOnSearchInput(event);
            }

            function setDataIsLinked(event) {
                const hash = event ? event.currentTarget.getAttribute("href").substring(1) : window.location.hash.substring(1);
                document.querySelectorAll('.diff-container').forEach((section) => {
                    if (section.id === hash) {
                        section.setAttribute("data-is-linked", "true");
                        if (!event) {
                            // Scroll to the linked element on page load
                            section.scrollIntoView({block: "start", inline: "nearest", behavior: "instant"});
                        }
                    } else {
                        section.setAttribute("data-is-linked", "false");
                    }
                });
            }

            function toggleFolders(open) {
                if (open) {
                    const toOpen = document.querySelectorAll('details.folder:open > ul > li > details.folder:not(:open)');
                    if (toOpen.length === 0) {
                        // We might need to open the root folders
                        document.querySelectorAll('.file-list > li > details.folder:not(:open)').forEach(d => d.open = true);
                    } else {
                        toOpen.forEach(d => d.open = true);
                    }
                } else {
                    document.querySelectorAll('details.folder:open').forEach(d => d.open = false);
                }
            }

            function toggleSidebar(show) {
                const sidebar = document.querySelector('.sidebar');
                const sidebarReveal = document.querySelector('.sidebar-reveal');
                const content = document.querySelector('.content');
                if (show) {
                    sidebar.style.display = 'block';
                    sidebarReveal.style.display = 'none';
                    content.style.padding = '20px';
                } else {
                    sidebar.style.display = 'none';
                    sidebarReveal.style.display = 'block';
                    content.style.padding = '20px 20px 20px 5px';
                }
            }

            function toggleMoreFileTree() {
                const moreOptions = document.querySelector('.file-tree-more');
                console.log(moreOptions.style.display);
                const show = moreOptions.style.display !== 'block';
                if (show) {
                    moreOptions.style.display = 'block';
                } else {
                    moreOptions.style.display = 'none';
                }
            }
        </script>
    </head>
    <body>
        <div class='container'>
            <div class='sidebar'>
                <div id="sidebar-contents">
                    <div class="search-area">
                        <div class="search-header">
                            <div>
                                <input type="search" class="search-field" id="search-include" placeholder="Include search..." oninput="onIncludeSearchInput(event)" />
                                <input type="search" class="search-field" id="search-exclude" placeholder="Exclude search..." oninput="onExcludeSearchInput(event)" />
                                <span id="searching_icon" style="display:none">...</span>
                            </div>

                            <button onclick="toggleSidebar(false)" title="Hide">
                                &#x2190;
                            </button>
                        </div>
                        <p>Showing <b id="file-count">{{ content|length }}</b> out of <b>{{ content|length }}</b> files</p>
                    </div>
                    <div class="file-tree">
                        <div class="file-tree-controls">
                            <div class="folder-collapse">
                                <button onclick="toggleFolders(true)" title="Expand current level">
                                    &#x02C4;
                                    <br/>
                                    &#x02C5;
                                </button>
                                <button onclick="toggleFolders(false)" title="Collapse all">
                                    &#x02C5;
                                    <br/>
                                    &#x02C4
                                </button>
                            </div>
                            <button onclick="toggleMoreFileTree()" title="Show more options"
                                class="file-tree-reveal-more">
                                    &#x22EE;
                            </button>
                        </div>
                        <div class="file-tree-more">
                            <h4>Show...</h4>
                            <div class="file-tree-more-option">
                                <input type="checkbox" id="show-old-files" checked
                                    onclick="debouncedOnSearchInput(event)"/>
                                <label for="show-old-files">Old files</label>
                            </div>

                            <div class="file-tree-more-option">
                                <input type="checkbox" id="show-new-files" checked
                                    onclick="debouncedOnSearchInput(event)"/>
                                <label for="show-new-files">New files</label>
                            </div>

                            <div class="file-tree-more-option">
                                <input type="checkbox" id="show-deleted-files" checked
                                    onclick="debouncedOnSearchInput(event)"/>
                                <label for="show-deleted-files">Deleted files</label>
                            </div>

                            <div class="file-tree-more-option">
                                <input type="checkbox" id="show-moved-files" checked
                                    onclick="debouncedOnSearchInput(event)"/>
                                <label for="show-moved-files">Moved files</label>
                            </div>
                        </div>
                        <ul class="file-list">
                            {{ render_sidebar_folder("", per_folder) }}
                        </ul>
                    </div>
                </div>
                <span id="empty_search" style="display:none">No results found</span>
            </div>
            <div class='sidebar-reveal'>
                <button onclick="toggleSidebar(true)" title="Show">
                    &#x2192;
                </button>
            </div>
            <div class='content'>
                <div class="diff-header">
                    <h2>Diff Report Between <b class="del">{{ old_reference.repr_notime() }}</b> And <b class="add">{{ new_reference.repr_notime() }}</b></h2>
                </div>
                <span id="empty_result" style="display:none">No matches</span>
                {{ render_diff_folder(per_folder) }}
            </div>
        </div>
    </body>
</html>
"""



import os
import platform
import subprocess
import textwrap

import pytest

from conan.test.utils.mocks import ConanFileMock
from conan.tools.env.environment import environment_wrap_command
from conan.test.utils.tools import TestClient, GenConanfile


@pytest.fixture()
def client():
    openssl = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            settings = "os"
            package_type = "shared-library"
            def package(self):
                with chdir(self, self.package_folder):
                    echo = "@echo off\necho MYOPENSSL={}!!".format(self.settings.os)
                    save(self, "bin/myopenssl.bat", echo)
                    save(self, "bin/myopenssl.sh", echo)
                    os.chmod("bin/myopenssl.sh", 0o777)
            """)

    cmake = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            settings = "os"
            requires = "openssl/1.0"
            def package(self):
                with chdir(self, self.package_folder):
                    echo = "@echo off\necho MYCMAKE={}!!".format(self.settings.os)
                    save(self, "mycmake.bat", echo + "\ncall myopenssl.bat")
                    save(self, "mycmake.sh", echo + "\n myopenssl.sh")
                    os.chmod("mycmake.sh", 0o777)

            def package_info(self):
                # Custom buildenv not defined by cpp_info
                self.buildenv_info.prepend_path("PATH", self.package_folder)
                self.buildenv_info.define("MYCMAKEVAR", "MYCMAKEVALUE!!")
            """)

    gtest = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            settings = "os"
            def package(self):
                with chdir(self, self.package_folder):
                    prefix = "@echo off\n" if self.settings.os == "Windows" else ""
                    echo = "{}echo MYGTEST={}!!".format(prefix, self.settings.os)
                    save(self, "bin/mygtest.bat", echo)
                    save(self, "bin/mygtest.sh", echo)
                    os.chmod("bin/mygtest.sh", 0o777)

            def package_info(self):
                self.runenv_info.define("MYGTESTVAR", "MyGTestValue{}".format(self.settings.os))
            """)
    client = TestClient()
    client.save({"cmake/conanfile.py": cmake,
                 "gtest/conanfile.py": gtest,
                 "openssl/conanfile.py": openssl})

    client.run("export openssl --name=openssl --version=1.0")
    client.run("export cmake --name=mycmake --version=1.0")
    client.run("export gtest --name=mygtest --version=1.0")

    myrunner_bat = "@echo off\necho MYGTESTVAR=%MYGTESTVAR%!!\n"
    myrunner_sh = "echo MYGTESTVAR=$MYGTESTVAR!!\n"
    client.save({"myrunner.bat": myrunner_bat,
                 "myrunner.sh": myrunner_sh}, clean_first=True)
    os.chmod(os.path.join(client.current_folder, "myrunner.sh"), 0o777)
    return client


@pytest.mark.parametrize("gtest_run_true", [True, False])
def test_complete(client, gtest_run_true):
    """By default, a test require has the run=False trait, so the PATH to the bat cannot be
    accessed"""
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile

        class Pkg(ConanFile):
            requires = "openssl/1.0"
            tool_requires = "mycmake/1.0"

            def build_requirements(self):
                {}

            def build(self):
                self.run("mycmake.bat", env="conanbuildenv")
                assert os.path.exists(os.path.join(self.generators_folder, "conanrunenv.sh"))
       """)
    if gtest_run_true:
        test_require = 'self.test_requires("mygtest/1.0", run=True)'
    else:
        test_require = 'self.test_requires("mygtest/1.0")'

    conanfile = conanfile.format(test_require)
    client.save({"conanfile.py": conanfile})
    client.run("install . -s:b os=Windows -s:h os=Linux --build=missing")
    # Run the BUILD environment
    if platform.system() == "Windows":
        cmd = environment_wrap_command(ConanFileMock(), "conanbuildenv", client.current_folder, "mycmake.bat")
        client.run_command(cmd)
        assert "MYCMAKE=Windows!!" in client.out
        assert "MYOPENSSL=Windows!!" in client.out

    # Run the RUN environment
    if platform.system() != "Windows":
        cmd = environment_wrap_command(ConanFileMock(), "conanrunenv", client.current_folder,
                                       "mygtest.sh && .{}myrunner.sh".format(os.sep))
        client.run_command(cmd, assert_error=not gtest_run_true)
        if gtest_run_true:
            assert "MYGTEST=Linux!!" in client.out
            assert "MYGTESTVAR=MyGTestValueLinux!!" in client.out

    if platform.system() == "Windows":
        client.run("build . -s:h os=Linux")
        assert "MYCMAKE=Windows!!" in client.out
        assert "MYOPENSSL=Windows!!" in client.out


def test_profile_included_multiple():
    client = TestClient()
    conanfile = textwrap.dedent("""\
        import os, platform
        from conan import ConanFile
        class Pkg(ConanFile):
            def generate(self):
                buildenv = self.buildenv.vars(self)
                self.output.info("MYVAR1: {}!!!".format(buildenv.get("MYVAR1")))
                self.output.info("MYVAR2: {}!!!".format(buildenv.get("MYVAR2")))
                self.output.info("MYVAR3: {}!!!".format(buildenv.get("MYVAR3")))
        """)

    myprofile = textwrap.dedent("""
       [buildenv]
       MYVAR1=MyVal1
       MYVAR3+=MyVal3
       """)
    other_profile = textwrap.dedent("""
       [buildenv]
       MYVAR1=MyValOther1
       MYVAR2=MyValOther2
       MYVAR3=MyValOther3
       """)
    client.save({"conanfile.py": conanfile,
                 "myprofile": myprofile,
                 "myprofile_include": "include(other_profile)\n" + myprofile,
                 "other_profile": other_profile})
    # The reference profile has priority
    client.run("install . -pr=myprofile_include")
    assert "MYVAR1: MyVal1!!!" in client.out
    assert "MYVAR2: MyValOther2!!!" in client.out
    assert "MYVAR3: MyValOther3 MyVal3!!!" in client.out

    # Equivalent to include is to put it first, then the last has priority
    client.run("install . -pr=other_profile -pr=myprofile")
    assert "MYVAR1: MyVal1!!!" in client.out
    assert "MYVAR2: MyValOther2!!!" in client.out
    assert "MYVAR3: MyValOther3 MyVal3!!!" in client.out


def test_profile_buildenv():
    client = TestClient()
    conanfile = textwrap.dedent("""\
        import os, platform
        from conan import ConanFile
        class Pkg(ConanFile):
            def generate(self):
                self.buildenv.vars(self).save_script("pkgenv")
                if platform.system() != "Windows":
                    os.chmod("pkgenv.sh", 0o777)

        """)
    # Some scripts in a random system folders, path adding to the profile [env]

    compiler_bat = "@echo off\necho MYCOMPILER!!\necho MYPATH=%PATH%"
    compiler_sh = "echo MYCOMPILER!!\necho MYPATH=$PATH"
    compiler2_bat = "@echo off\necho MYCOMPILER2!!\necho MYPATH2=%PATH%"
    compiler2_sh = "echo MYCOMPILER2!!\necho MYPATH2=$PATH"

    myprofile = textwrap.dedent("""
           [buildenv]
           PATH+=(path){}
           mypkg*:PATH=!
           mypkg*:PATH+=(path){}
           """.format(os.path.join(client.current_folder, "compiler"),
                      os.path.join(client.current_folder, "compiler2")))
    client.save({"conanfile.py": conanfile,
                 "myprofile": myprofile,
                 "compiler/mycompiler.bat": compiler_bat,
                 "compiler/mycompiler.sh": compiler_sh,
                 "compiler2/mycompiler.bat": compiler2_bat,
                 "compiler2/mycompiler.sh": compiler2_sh})

    os.chmod(os.path.join(client.current_folder, "compiler", "mycompiler.sh"), 0o777)
    os.chmod(os.path.join(client.current_folder, "compiler2", "mycompiler.sh"), 0o777)

    client.run("install . -pr=myprofile")
    # Run the BUILD environment
    ext = "bat" if platform.system() == "Windows" else "sh"  # TODO: Decide on logic .bat vs .sh
    cmd = environment_wrap_command(ConanFileMock(), "conanbuildenv", client.current_folder,
                                   "mycompiler.{}".format(ext))
    client.run_command(cmd)
    assert "MYCOMPILER!!" in client.out
    assert "MYPATH=" in client.out

    # Now with pkg-specific env-var
    client.run("install . --name=mypkg --version=1.0 -pr=myprofile")
    client.run_command(cmd)
    assert "MYCOMPILER2!!" in client.out
    assert "MYPATH2=" in client.out


def test_transitive_order():
    # conanfile.py -(br)-> cmake -> openssl (unknown=static)
    #     \                    \-(br)-> gcc
    #      \--------(br)-> gcc
    #       \---------------------> openssl
    gcc = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            def package_info(self):
                self.runenv_info.append("MYVAR", "MyGCCValue")
        """)
    openssl = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            tool_requires = "gcc/1.0"
            package_type = "shared-library"
            def package_info(self):
                self.runenv_info.append("MYVAR", "MyOpenSSL{}Value".format(self.settings.os))
        """)
    cmake = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            requires = "openssl/1.0"
            tool_requires = "gcc/1.0"
            def package_info(self):
                self.runenv_info.append("MYVAR", "MyCMakeRunValue")
                self.buildenv_info.append("MYVAR", "MyCMakeBuildValue")
        """)
    client = TestClient()
    client.save({"gcc/conanfile.py": gcc,
                 "cmake/conanfile.py": cmake,
                 "openssl/conanfile.py": openssl})

    client.run("export gcc --name=gcc --version=1.0")
    client.run("export openssl --name=openssl --version=1.0")
    client.run("export cmake --name=cmake --version=1.0")

    consumer = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.env import VirtualBuildEnv, VirtualRunEnv
        class Pkg(ConanFile):
            requires = "openssl/1.0"
            tool_requires = "cmake/1.0", "gcc/1.0"
            def generate(self):
                buildenv = VirtualBuildEnv(self).vars()
                self.output.info("BUILDENV: {}!!!".format(buildenv.get("MYVAR")))
                runenv = VirtualRunEnv(self).vars()
                self.output.info("RUNENV: {}!!!".format(runenv.get("MYVAR")))
        """)
    client.save({"conanfile.py": consumer}, clean_first=True)
    client.run("install . -s:b os=Windows -s:h os=Linux --build='*'")
    assert "BUILDENV: MyOpenSSLWindowsValue MyGCCValue "\
           "MyCMakeRunValue MyCMakeBuildValue!!!" in client.out
    assert "RUNENV: MyOpenSSLLinuxValue!!!" in client.out

    # Even if the generator is duplicated in command line (it used to fail due to bugs)
    client.run("install . -s:b os=Windows -s:h os=Linux --build='*' -g VirtualRunEnv -g VirtualBuildEnv")
    assert "BUILDENV: MyOpenSSLWindowsValue MyGCCValue "\
           "MyCMakeRunValue MyCMakeBuildValue!!!" in client.out
    assert "RUNENV: MyOpenSSLLinuxValue!!!" in client.out


def test_buildenv_from_requires():
    openssl = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def package_info(self):
                self.buildenv_info.append("OpenSSL_ROOT",
                                          "MyOpenSSL{}Value".format(self.settings.os))
        """)
    poco = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            requires = "openssl/1.0"
            settings = "os"
            def package_info(self):
                self.buildenv_info.append("Poco_ROOT", "MyPoco{}Value".format(self.settings.os))
        """)
    client = TestClient()
    client.save({"poco/conanfile.py": poco,
                 "openssl/conanfile.py": openssl})

    client.run("export openssl --name=openssl --version=1.0")
    client.run("export poco --name=poco --version=1.0")

    consumer = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.env import VirtualBuildEnv
        class Pkg(ConanFile):
            requires = "poco/1.0"
            def generate(self):
                buildenv = VirtualBuildEnv(self).vars()
                self.output.info("BUILDENV POCO: {}!!!".format(buildenv.get("Poco_ROOT")))
                self.output.info("BUILDENV OpenSSL: {}!!!".format(buildenv.get("OpenSSL_ROOT")))
        """)
    client.save({"conanfile.py": consumer}, clean_first=True)
    client.run("install . -s:b os=Windows -s:h os=Linux --build='*' -g VirtualBuildEnv")
    assert "BUILDENV POCO: MyPocoLinuxValue!!!" in client.out
    assert "BUILDENV OpenSSL: MyOpenSSLLinuxValue!!!" in client.out


def test_diamond_repeated():
    pkga = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            def package_info(self):
                self.runenv_info.define("MYVAR1", "PkgAValue1")
                self.runenv_info.append("MYVAR2", "PkgAValue2")
                self.runenv_info.prepend("MYVAR3", "PkgAValue3")
                self.runenv_info.prepend("MYVAR4", "PkgAValue4")
        """)
    pkgb = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            requires = "pkga/1.0"
            def package_info(self):
                self.runenv_info.append("MYVAR1", "PkgBValue1")
                self.runenv_info.append("MYVAR2", "PkgBValue2")
                self.runenv_info.prepend("MYVAR3", "PkgBValue3")
                self.runenv_info.prepend("MYVAR4", "PkgBValue4")
        """)
    pkgc = textwrap.dedent(r"""
        from conan import ConanFile
        class Pkg(ConanFile):
            requires = "pkga/1.0"
            def package_info(self):
                self.runenv_info.append("MYVAR1", "PkgCValue1")
                self.runenv_info.append("MYVAR2", "PkgCValue2")
                self.runenv_info.prepend("MYVAR3", "PkgCValue3")
                self.runenv_info.prepend("MYVAR4", "PkgCValue4")
        """)
    pkgd = textwrap.dedent(r"""
       from conan import ConanFile
       class Pkg(ConanFile):
           requires = "pkgb/1.0", "pkgc/1.0"
           def package_info(self):
               self.runenv_info.append("MYVAR1", "PkgDValue1")
               self.runenv_info.append("MYVAR2", "PkgDValue2")
               self.runenv_info.prepend("MYVAR3", "PkgDValue3")
               self.runenv_info.define("MYVAR4", "PkgDValue4")
       """)
    pkge = textwrap.dedent(r"""
       from conan import ConanFile
       from conan.tools.env import VirtualRunEnv
       class Pkg(ConanFile):
           requires = "pkgd/1.0"
           def generate(self):
                env = VirtualRunEnv(self)
                runenv = env.vars(scope="run")
                self.output.info("MYVAR1: {}!!!".format(runenv.get("MYVAR1")))
                self.output.info("MYVAR2: {}!!!".format(runenv.get("MYVAR2")))
                self.output.info("MYVAR3: {}!!!".format(runenv.get("MYVAR3")))
                self.output.info("MYVAR4: {}!!!".format(runenv.get("MYVAR4")))
                env.generate()
       """)
    client = TestClient()
    client.save({"pkga/conanfile.py": pkga,
                 "pkgb/conanfile.py": pkgb,
                 "pkgc/conanfile.py": pkgc,
                 "pkgd/conanfile.py": pkgd,
                 "pkge/conanfile.py": pkge})

    client.run("export pkga --name=pkga --version=1.0")
    client.run("export pkgb --name=pkgb --version=1.0")
    client.run("export pkgc --name=pkgc --version=1.0")
    client.run("export pkgd --name=pkgd --version=1.0")

    client.run("install pkge --build=missing")
    # PkgB has higher priority (included first) so it is appended last and prepended first (wrtC)
    assert "MYVAR1: PkgAValue1 PkgCValue1 PkgBValue1 PkgDValue1!!!" in client.out
    assert "MYVAR2: PkgAValue2 PkgCValue2 PkgBValue2 PkgDValue2!!!" in client.out
    assert "MYVAR3: PkgDValue3 PkgBValue3 PkgCValue3 PkgAValue3!!!" in client.out
    assert "MYVAR4: PkgDValue4!!!" in client.out

    # No settings always sh
    conanrun = client.load("pkge/conanrunenv.sh")
    assert "PATH" not in conanrun
    assert 'export MYVAR1="PkgAValue1 PkgCValue1 PkgBValue1 PkgDValue1"' in conanrun
    assert 'export MYVAR2="$MYVAR2 PkgAValue2 PkgCValue2 PkgBValue2 PkgDValue2"' in conanrun
    assert 'export MYVAR3="PkgDValue3 PkgBValue3 PkgCValue3 PkgAValue3 $MYVAR3"' in conanrun
    assert 'export MYVAR4="PkgDValue4"' in conanrun


@pytest.mark.parametrize("require_run", [True, False])
def test_environment_scripts_generated_envvars(require_run):
    """If the regular require doesn't declare the 'run' trait, the conanrunenv won't be there,
    unless for example, the require_pkg has the pkg_type to Application"""
    consumer_pkg = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.env import VirtualBuildEnv, VirtualRunEnv
        class Pkg(ConanFile):
            settings = "os"
            requires = "require_pkg/1.0"
            tool_requires = "build_require_pkg/1.0"
            generators = "VirtualRunEnv", "VirtualBuildEnv"
        """)

    client = TestClient()
    conanfile_br = (GenConanfile().with_package_file("bin/myapp", "myexe")
                                  .with_package_file("lib/mylib", "mylibcontent")
                                  .with_settings("os"))
    conanfile_require = (GenConanfile().with_package_file("bin/myapp", "myexe")
                                       .with_package_file("lib/mylib", "mylibcontent")
                                       .with_settings("os"))
    if require_run:
        conanfile_require.with_package_type("application")
    client.save({"build_require_pkg/conanfile.py": conanfile_br,
                 "require_pkg/conanfile.py": conanfile_require,
                 "consumer_pkg/conanfile.py": consumer_pkg})

    client.run("export build_require_pkg --name=build_require_pkg --version=1.0")
    client.run("export require_pkg --name=require_pkg --version=1.0")

    client.run("install consumer_pkg --build='*'")
    if platform.system() == "Windows":
        conanbuildenv = client.load("consumer_pkg/conanbuildenv.bat")
        if require_run:
            conanrunenv = client.load("consumer_pkg/conanrunenv.bat")
            assert "LD_LIBRARY_PATH" not in conanbuildenv
            assert "LD_LIBRARY_PATH" not in conanrunenv
        else:
            assert not os.path.exists("consumer_pkg/conanrunenv.bat")
    else:
        if require_run:
            conanbuildenv = client.load("consumer_pkg/conanbuildenv.sh")
            conanrunenv = client.load("consumer_pkg/conanrunenv.sh")
            assert "LD_LIBRARY_PATH" in conanbuildenv
            assert "LD_LIBRARY_PATH" in conanrunenv
        else:
            assert not os.path.exists("consumer_pkg/conanrunenv.sh")

    if require_run:
        # Build context LINUX - Host context LINUX
        client.run("install consumer_pkg -s:b os=Linux -s:h os=Linux --build='*'")
        conanbuildenv = client.load("consumer_pkg/conanbuildenv.sh")
        conanrunenv = client.load("consumer_pkg/conanrunenv.sh")
        assert "LD_LIBRARY_PATH" in conanbuildenv
        assert "LD_LIBRARY_PATH" in conanrunenv

        # Build context WINDOWS - Host context WINDOWS
        client.run("install consumer_pkg -s:b os=Windows -s:h os=Windows --build='*'")
        conanbuildenv = client.load("consumer_pkg/conanbuildenv.bat")
        conanrunenv = client.load("consumer_pkg/conanrunenv.bat")
        assert "LD_LIBRARY_PATH" not in conanbuildenv
        assert "LD_LIBRARY_PATH" not in conanrunenv

        # Build context LINUX - Host context WINDOWS
        client.run("install consumer_pkg -s:b os=Linux -s:h os=Windows --build='*'")
        conanbuildenv = client.load("consumer_pkg/conanbuildenv.sh")
        conanrunenv = client.load("consumer_pkg/conanrunenv.bat")
        assert "LD_LIBRARY_PATH" in conanbuildenv
        assert "LD_LIBRARY_PATH" not in conanrunenv

        # Build context WINDOWS - Host context LINUX
        client.run("install consumer_pkg -s:b os=Windows -s:h os=Linux --build='*'")
        conanbuildenv = client.load("consumer_pkg/conanbuildenv.bat")
        conanrunenv = client.load("consumer_pkg/conanrunenv.sh")
        assert "LD_LIBRARY_PATH" not in conanbuildenv
        assert "LD_LIBRARY_PATH" in conanrunenv


@pytest.mark.parametrize("deactivation_mode", ["function", None])
def test_multiple_deactivate(deactivation_mode):
    conanfile = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.env import Environment
        class Pkg(ConanFile):
            def generate(self):
                e1 = Environment()
                e1.define("VAR1", "Value1")
                e1.vars(self).save_script("mybuild1")
                e2 = Environment()
                e2.define("VAR2", "Value2")
                e2.vars(self).save_script("mybuild2")
        """)
    display_bat = textwrap.dedent("""\
        @echo off
        echo VAR1=%VAR1%!!
        echo VAR2=%VAR2%!!
        """)
    display_sh = textwrap.dedent("""\
        echo VAR1=$VAR1!!
        echo VAR2=$VAR2!!
        """)
    client = TestClient()
    client.save({"conanfile.py": conanfile,
                 "display.bat": display_bat,
                 "display.sh": display_sh})
    os.chmod(os.path.join(client.current_folder, "display.sh"), 0o777)
    client.run(f"install . {f'-c=tools.env:deactivation_mode={deactivation_mode}' if deactivation_mode else ''} ")

    for _ in range(2):  # Just repeat it, so we can check things keep working
        if platform.system() == "Windows":
            cmd = "conanbuild.bat && display.bat && deactivate_conanbuild.bat && display.bat"
        else:
            deactivate_cmd = "deactivate_conanbuild" if deactivation_mode else ". ./deactivate_conanbuild.sh"
            cmd = f'. ./conanbuild.sh && ./display.sh && {deactivate_cmd} && ./display.sh'
        out, _ = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                  shell=True, cwd=client.current_folder).communicate()
        out = out.decode()
        assert "VAR1=Value1!!" in out
        assert "VAR2=Value2!!" in out
        assert 3 == str(out).count("Restoring environment")
        assert "VAR1=!!" in out
        assert "VAR2=!!" in out


@pytest.mark.parametrize("deactivation_mode", ["function", None])
def test_multiple_deactivate_order(deactivation_mode):
    """
    https://github.com/conan-io/conan/issues/13693
    """
    conanfile = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.env import Environment
        class Pkg(ConanFile):
            def generate(self):
                e1 = Environment()
                e1.define("MYVAR", "Value1")
                e1.vars(self).save_script("mybuild1")
                e2 = Environment()
                e2.define("MYVAR", "Value2")
                e2.vars(self).save_script("mybuild2")
        """)
    display_bat = textwrap.dedent("""\
        @echo off
        echo MYVAR=%MYVAR%!!
        """)
    display_sh = textwrap.dedent("""\
        echo MYVAR=$MYVAR!!
        """)
    client = TestClient()
    client.save({"conanfile.py": conanfile,
                 "display.bat": display_bat,
                 "display.sh": display_sh})
    os.chmod(os.path.join(client.current_folder, "display.sh"), 0o777)
    client.run(f"install . {f'-c=tools.env:deactivation_mode={deactivation_mode}' if deactivation_mode else ''} ")

    for _ in range(2):  # Just repeat it, so we can check things keep working
        if platform.system() == "Windows":
            cmd = "conanbuild.bat && display.bat && deactivate_conanbuild.bat && display.bat"
        else:
            deactivate_cmd = "deactivate_conanbuild" if deactivation_mode else ". ./deactivate_conanbuild.sh"
            cmd = f'. ./conanbuild.sh && ./display.sh && {deactivate_cmd} && ./display.sh'
        out, _ = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                  shell=True, cwd=client.current_folder).communicate()
        out = out.decode()
        assert "MYVAR=Value2!!" in out
        assert 3 == str(out).count("Restoring environment")
        assert "MYVAR=!!" in out


@pytest.mark.skipif(platform.system() == "Windows", reason="Shell script test")
@pytest.mark.parametrize("deactivation_mode", ["function", None])
def test_deactivate_missing_vars_stay_missing(deactivation_mode):
    """ Tests that these two cases preserve variable status
    1.
    export FOO=
    ./conanrunenv.sh that changes FOO to something with a value
    ./deactivate_conanrunenv.sh
    FOO is still empty

    2.
    BAR is not defined
    ./conanrunenv.sh that changes BAR to something with a value
    ./deactivate_conanrunenv.sh
    BAR is still not defined

    3.
    BAZ is defined to some value
    ./conanrunenv.sh that unsets BAZ
    ./deactivate_conanrunenv.sh
    BAZ is still defined to some value

    4.
    FOOBAR is empty
    ./conanrunenv.sh that unsets FOOBAR
    ./deactivate_conanrunenv.sh
    FOOBAR is still empty
    """
    conanfile = textwrap.dedent(r"""
            from conan import ConanFile
            from conan.tools.env import Environment
            class Pkg(ConanFile):
                def generate(self):
                    e1 = Environment()
                    e1.define("FOO", "Value1")
                    e1.define("BAR", "Value2")
                    e1.unset("BAZ")
                    e1.unset("FOOBAR")
                    e1.vars(self).save_script("mybuild1")
            """)
    display_sh = textwrap.dedent("""\
            echo FOO=$FOO!!
            if [ -n "${BAR+x}" ]; then echo "BAR EXISTS!!"; fi;
            echo BAZ=$BAZ!!
            echo FOOBAR=$FOOBAR!!
            """)
    client = TestClient(light=True)
    client.save({"conanfile.py": conanfile,
                 "display.sh": display_sh})
    os.chmod(os.path.join(client.current_folder, "display.sh"), 0o777)
    client.run(f"install . {f'-c=tools.env:deactivation_mode={deactivation_mode}' if deactivation_mode else ''} ")

    deactivate_cmd = "deactivate_conanbuild" if deactivation_mode else ". ./deactivate_conanbuild.sh"
    cmd = (f'export FOO=&& export BAZ=Value3 && export FOOBAR='
           f'&& . ./conanbuild.sh && {deactivate_cmd} && ./display.sh')
    out, _ = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                              shell=True, cwd=client.current_folder).communicate()
    out = out.decode()
    assert "FOO=!!" in out
    assert "BAR EXISTS!!" not in out
    assert "BAZ=Value3!!" in out
    assert "FOOBAR=!!" in out


@pytest.mark.skipif(platform.system() != "Windows", reason="Path problem in Windows only")
@pytest.mark.parametrize("num_deps", [3, ])
def test_massive_paths(num_deps):
    """ This test proves that having too many dependencies that will result in a very long PATH
    env-var in the consumer by one VirtualXXXEnv environment, will overflow.
    https://github.com/conan-io/conan/issues/9565
    This seems an unsolvable limitation, the only alternatives are:
    - shorten the paths in general (shorter cache paths)
    - add exclusively the paths of needed things (better visibility)
    Seems that Conan 2.0 will improve over these things, allowing larger dependencies graphs without
    failing. Besides that, it might use the deployers to workaround shared-libs running scenarios.

    The test is parameterized for being fast an passing, but if we add a num_deps >= 80 approx,
    it will start to enter the failing scenarios. Not adding the >=80 scenario, because that tests
    takes 1 minute by itself, not worth the value.
    """
    client = TestClient(path_with_spaces=False)
    compiler_bat = "@echo off\necho MYTOOL {}!!\n"
    conanfile = textwrap.dedent("""\
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        class Pkg(ConanFile):
            exports_sources = "*"
            package_type = "application"
            def package(self):
                copy(self, "*", self.build_folder, os.path.join(self.package_folder, "bin"))
        """)

    for i in range(num_deps):
        client.save({"conanfile.py": conanfile,
                     "mycompiler{}.bat".format(i): compiler_bat.format(i)})
        client.run("create . --name=pkg{} --version=0.1".format(i))

    conanfile = textwrap.dedent("""\
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            requires = {}
            generators = "VirtualRunEnv"
        """)
    requires = ", ".join('"pkg{}/0.1"'.format(i) for i in range(num_deps))
    conanfile = conanfile.format(requires)
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("install . -c tools.env.virtualenv:powershell=True")
    assert os.path.isfile(os.path.join(client.current_folder, "conanrunenv.ps1"))
    assert not os.path.isfile(os.path.join(client.current_folder, "conanrunenv.bat"))
    for i in range(num_deps):
        cmd = environment_wrap_command(ConanFileMock(), "conanrunenv", client.current_folder,
                                       "mycompiler{}.bat".format(i))
        # if num_deps > 50:  # to be safe if we change the "num_deps" number
        #    client.run_command(cmd, assert_error=True)
        #    assert "is not recognized as an internal" in client.out
        client.run_command(cmd)
        assert "MYTOOL {}!!".format(i) in client.out

    # Test .bats now
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("install .")
    assert not os.path.isfile(os.path.join(client.current_folder, "conanrunenv.ps1"))
    assert os.path.isfile(os.path.join(client.current_folder, "conanrunenv.bat"))
    for i in range(num_deps):
        cmd = environment_wrap_command(ConanFileMock(), "conanrunenv", client.current_folder,
                                       "mycompiler{}.bat".format(i))
        # if num_deps > 50:  # to be safe if we change the "num_deps" number
        #    client.run_command(cmd, assert_error=True)
        #    # This also fails, but without an error message (in my terminal, it kills the terminal!)
        # else:
        client.run_command(cmd)
        assert "MYTOOL {}!!".format(i) in client.out


@pytest.mark.parametrize("deactivation_mode", ["function", None])
def test_profile_build_env_spaces(deactivation_mode):
    display_bat = textwrap.dedent("""\
        @echo off
        echo VAR1=%VAR1%!!
        """)
    display_sh = textwrap.dedent("""\
        echo VAR1=$VAR1!!
        """)
    client = TestClient()
    client.save({"conanfile.txt": "",
                 "profile": "[buildenv]\nVAR1 = VALUE1",
                 "display.bat": display_bat,
                 "display.sh": display_sh})
    os.chmod(os.path.join(client.current_folder, "display.sh"), 0o777)
    client.run(f"install . -g VirtualBuildEnv -pr=profile {f'-c=tools.env:deactivation_mode={deactivation_mode}' if deactivation_mode else ''} ")

    if platform.system() == "Windows":
        cmd = "conanbuild.bat && display.bat && deactivate_conanbuild.bat && display.bat"
    else:
        deactivate_cmd = "deactivate_conanbuild" if deactivation_mode else ". ./deactivate_conanbuild.sh"
        cmd = f'. ./conanbuild.sh && ./display.sh && {deactivate_cmd} && ./display.sh'
    out, _ = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                              shell=True, cwd=client.current_folder).communicate()
    out = out.decode()
    assert "VAR1= VALUE1!!" in out
    assert "Restoring environment" in out
    assert "VAR1=!!" in out


def test_deactivate_location():
    conanfile = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.env import Environment
        class Pkg(ConanFile):
            def package_info(self):
                self.buildenv_info.define("FOO", "BAR")
        """)
    client = TestClient()
    client.save({"pkg.py": conanfile})
    client.run("create pkg.py --name pkg --version 1.0")
    client.run("install --requires pkg/1.0@ -g VirtualBuildEnv -of=myfolder -s build_type=Release -s arch=x86_64")

    source_cmd, script_ext = ("myfolder\\", ".bat") if platform.system() == "Windows" else (". ./myfolder/", ".sh")
    cmd = "{}conanbuild{}".format(source_cmd, script_ext)

    subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True,
                     cwd=client.current_folder).communicate()

    assert not os.path.exists(os.path.join(client.current_folder,
                                           "deactivate_conanbuildenv-release-x86_64{}".format(script_ext)))

    assert os.path.exists(os.path.join(client.current_folder, "myfolder",
                                       "deactivate_conanbuildenv-release-x86_64{}".format(script_ext)))


@pytest.mark.skipif(platform.system() == "Windows", reason="Requires sh")
def test_skip_virtualbuildenv_run():
    # Build require
    conanfile = textwrap.dedent(r"""
           from conan import ConanFile
           class Pkg(ConanFile):
               def package_info(self):
                   self.buildenv_info.define("FOO", "BAR")
           """)
    client = TestClient()
    client.save({"pkg.py": conanfile})
    client.run("create pkg.py --name pkg --version 1.0")

    # consumer
    conanfile = textwrap.dedent(r"""
               import os
               from conan import ConanFile
               class Consumer(ConanFile):
                   tool_requires = "pkg/1.0"
                   exports_sources = "my_script.sh"
                   # This can be removed at Conan 2
                   generators = "VirtualBuildEnv"
                   def build(self):
                       path = os.path.join(self.source_folder, "my_script.sh")
                       os.chmod(path, 0o777)
                       self.run("'{}'".format(path))
               """)
    my_script = 'echo FOO is $FOO'
    client.save({"conanfile.py": conanfile, "my_script.sh": my_script})
    client.run("create . --name consumer --version 1.0")
    assert "FOO is BAR" in client.out

    # If we pass env=None no "conanbuild" is applied
    # self.run("'{}'".format(path), env=None)
    conanfile = conanfile.replace(".format(path))",
                                  ".format(path), env=None)")
    client.save({"conanfile.py": conanfile})
    client.run("create . --name consumer --version 1.0")
    assert "FOO is BAR" not in client.out


def test_files_always_created():
    """ test that even if there are no env-variables, the generators always create files,
    they will be mostly empty, but exist
    """
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "consumer/conanfile.txt": "[requires]\ndep/0.1"})
    c.run("create dep")
    c.run("install consumer -g VirtualBuildEnv -g VirtualRunEnv -of=.")
    ext = "bat" if platform.system() == "Windows" else "sh"

    arch = c.get_default_host_profile().settings['arch']
    assert os.path.isfile(os.path.join(c.current_folder, f"conanbuild.{ext}"))
    assert os.path.isfile(os.path.join(c.current_folder, f"conanrun.{ext}"))
    assert os.path.isfile(os.path.join(c.current_folder, f"conanbuildenv-release-{arch}.{ext}"))
    assert os.path.isfile(os.path.join(c.current_folder, f"conanbuildenv-release-{arch}.{ext}"))


def test_error_with_dots_virtualenv():
    # https://github.com/conan-io/conan/issues/12163
    tool = textwrap.dedent(r"""
        from conan import ConanFile
        class ToolConan(ConanFile):
            name = "tool"
            version = "0.1"
            settings = "arch", "os"

            def package_info(self):
                self.buildenv_info.define("DUMMY", "123456")
        """)
    test_package = textwrap.dedent(r"""
        from conan import ConanFile
        import os

        class ToolTestConan(ConanFile):
            name = "app"
            version = "0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "VirtualBuildEnv"

            def build_requirements(self):
                self.tool_requires("tool/0.1")

            def build(self):
                self.run("echo DUMMY=$DUMMY")
                self.run("set")
            """)
    client = TestClient()
    client.save({"dep/conanfile.py": tool,
                 "consumer/conanfile.py": test_package})

    client.run("create dep -s:b arch=armv8.3")
    client.run("create consumer -s arch=armv8.3")
    assert "DUMMY=123456" in client.out


def test_runenv_info_propagated():
    """
    A runenv_info in a recipe, which is required by an executable that is used
    in another place as a tool_require (also its own test_package), should be
    propagated
    test_package --(tool_requires)->tools-(requires)->lib(defines runenv_info)
    https://github.com/conan-io/conan/issues/12939
    """
    c = TestClient()
    # NOTE: The lib contains ``package_type = "shared-library"`` to force propagation of runenv_info
    lib = textwrap.dedent("""
        from conan import ConanFile
        class Lib(ConanFile):
            name = "lib"
            version = "0.1"
            settings = "build_type"
            package_type = "shared-library"
            def package_info(self):
                self.runenv_info.define("MYLIBVAR", f"MYLIBVALUE:{self.settings.build_type}")
        """)
    tool_test_package = textwrap.dedent("""
        import platform
        from conan import ConanFile
        class TestTool(ConanFile):
            settings = "build_type"
            test_type = "explicit"
            generators = "VirtualBuildEnv"
            def build_requirements(self):
                self.tool_requires(self.tested_reference_str)
            def build(self):
                self.output.info(f"Building TEST_PACKAGE IN {self.settings.build_type}!!")
                if platform.system()!= "Windows":
                    self.run("echo MYLIBVAR=$MYLIBVAR")
                else:
                    self.run("set MYLIBVAR")
            def test(self):
                pass
        """)
    c.save({"lib/conanfile.py": lib,
            "tool/conanfile.py": GenConanfile("tool", "0.1").with_requires("lib/0.1"),
            "tool/test_package/conanfile.py": tool_test_package})
    c.run("create lib -s build_type=Release")
    c.run("create tool --build-require -s:b build_type=Release -s:h build_type=Debug")
    assert "tool/0.1 (test package): Building TEST_PACKAGE IN Debug!!" in c.out
    assert "MYLIBVAR=MYLIBVALUE:Release" in c.out


def test_deactivate_relocatable_substitute():
    c = TestClient()
    # this cannot be tested in CI, because permissions over root folder
    # c.current_folder = "/build"
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("install . -s os=Linux -s:b os=Linux")
    conanbuild = c.load("conanbuildenv.sh")
    result = os.path.join("$script_folder", "deactivate_conanbuildenv.sh")
    assert f'"{result}"' in conanbuild


class TestDotEnv:
    def test_dotenv(self):
        c = TestClient()
        other_path = os.path.join(c.current_folder, "my", "rel", "path")
        myprofile = textwrap.dedent(f"""
            [buildenv]
            MYVAR1=MyVal1
            MYVAR3+=MyVal3
            MYPATH+=(path)/some/path/here
            MYOTHER_PATH=+(path){other_path}

            [runenv]
            MYRUNVAR=SomeVal1

            [conf]
            tools.env:dotenv=True
            """)
        c.save({"conanfile.txt": "",
                "myprofile": myprofile})
        c.run("install . -pr=myprofile -s build_type=Release")
        dotenv = c.load("conanbuildenv-Release.env")
        expected = os.path.join(c.current_folder, "my", "rel", "path")
        assert f'MYOTHER_PATH="{expected}"' in dotenv
        assert f'MYPATH="/some/path/here"' in dotenv
        assert 'MYVAR3="MyVal3"' in dotenv
        assert 'MYVAR1="MyVal1"' in dotenv
        dotenv = c.load("conanrunenv-Release.env")
        assert f'MYRUNVAR="SomeVal1"' in dotenv

    def test_generate_only_dotenv(self):
        # If for some reason a recipe only wants the dot env files, doable
        c = TestClient()
        myprofile = textwrap.dedent(f"""
            [buildenv]
            MYVAR1=MyVal1
            [runenv]
            MYRUNVAR=SomeVal1
            """)
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.env import VirtualBuildEnv, VirtualRunEnv
            class Lib(ConanFile):
                def generate(self):
                    VirtualBuildEnv(self).vars().save_dotenv("mybuild.env")
                    VirtualRunEnv(self)
            """)
        c.save({"conanfile.py": conanfile,
                "myprofile": myprofile})
        c.run("install . -pr=myprofile")
        assert set(os.listdir(c.current_folder)) == {'conanfile.py', 'mybuild.env', 'myprofile'}
        assert 'MYVAR1="MyVal1"' in c.load("mybuild.env")



import sqlite3
import threading
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from typing import Tuple, List


class BaseDbTable:
    table_name: str = None
    columns_description: List[Tuple[str, type]] = None
    row_type: namedtuple = None
    columns: namedtuple = None
    unique_together: tuple = None
    _lock: threading.Lock = None
    _lock_storage = defaultdict(threading.Lock)

    def __init__(self, filename):
        self.filename = filename
        column_names: List[str] = [it[0] for it in self.columns_description]
        self.row_type = namedtuple('_', column_names)
        self.columns = self.row_type(*column_names)
        self._lock = self._lock_storage[self.filename]

    @contextmanager
    def db_connection(self):
        assert self._lock.acquire(timeout=20), "Conan failed to acquire database lock"
        connection = sqlite3.connect(self.filename, isolation_level=None, timeout=20)
        try:
            yield connection
        finally:
            connection.close()
            self._lock.release()

    def create_table(self):
        def field(name, typename, nullable=False, unique=False):
            field_str = name
            if typename is str:
                field_str += ' text'
            elif typename is int:
                field_str += ' integer'
            else:
                assert typename is float, f"sqlite3 type not mapped for type '{typename}'"
                field_str += ' real'

            if not nullable:
                field_str += ' NOT NULL'

            if unique:
                field_str += ' UNIQUE'

            return field_str

        fields = ', '.join([field(*it) for it in self.columns_description])
        guard = 'IF NOT EXISTS'
        table_checks = f", UNIQUE({', '.join(self.unique_together)})" if self.unique_together else ''
        with self.db_connection() as conn:
            conn.execute(f"CREATE TABLE {guard} {self.table_name} ({fields} {table_checks});")



from conans.migrations import Migrator


class ServerMigrator(Migrator):

    def __init__(self, conf_path, store_path, current_version, force_migrations):
        self.force_migrations = force_migrations
        self.store_path = store_path
        super(ServerMigrator, self).__init__(conf_path, current_version)



import textwrap
from string import Template

import pytest

from conan.test.utils.tools import TestClient


def test_configure_args():
    client = TestClient()

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"
            def build(self):
                cmake = CMake(self)
                cmake.configure(variables={"MYVAR": "MYVALUE"})
                cmake.build(cli_args=["--verbosebuild"], build_tool_args=["-something"])
                cmake.test(cli_args=["--testverbose"], build_tool_args=["-testok"])

            def run(self, *args, **kwargs):
                self.output.info("MYRUN: {}".format(*args))
            """)
    client.save({"conanfile.py": conanfile})
    client.run("create . ")
    # TODO: This check is ugly, because the command line is so different in each platform,
    #  and args_to_string() is doing crazy stuff
    assert '-DMYVAR="MYVALUE"' in client.out
    assert "--verbosebuild" in client.out
    assert "-something" in client.out
    assert "--testverbose" in client.out
    assert "-testok" in client.out


@pytest.mark.parametrize("argument, output_args", [
    ("'target'", "--target target"),
    ("['target']", "--target target"),
    ("['target1', 'target2']", "--target target1 target2"),
])
def test_multiple_targets(argument, output_args):
    client = TestClient()

    conanfile_template = Template(textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake
        class Pkg(ConanFile):
            generators = "CMakeToolchain"
            name = "pkg"
            version = "0.1"
            settings = "os", "compiler", "build_type", "arch"
            def build(self):
                cmake = CMake(self)
                cmake.build(target=${argument})

            def run(self, *args, **kwargs):
                self.output.info("MYRUN: {}".format(*args))
            """))
    conanfile = conanfile_template.substitute(argument=argument)
    client.save({"conanfile.py": conanfile})
    client.run("create . ")
    assert output_args in client.out



import os
import platform
import tarfile
import zipfile

import pytest
import sys

from conan.internal.api.uploader import gzopen_without_timestamps
from conan.tools.files import load
from conan.tools.files.files import untargz, unzip
from conan.errors import ConanException
from conan.internal.model.manifest import gather_files
from conan.test.utils.mocks import ConanFileMock
from conan.test.utils.mocks import RedirectedTestOutput
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import redirect_output
from conan.internal.util.files import chdir, save
from conan.internal.util.files import rmdir


class TestZipExtractPlain:

    def _zipdir(self, path, zipfilename, folder_entry=None):
        with zipfile.ZipFile(zipfilename, 'w', zipfile.ZIP_DEFLATED) as z:
            if folder_entry:
                zif = zipfile.ZipInfo(folder_entry + "/")
                z.writestr(zif, "")

            for root, _, files in os.walk(path):
                for f in files:
                    file_path = os.path.join(root, f)
                    if file_path == zipfilename:
                        continue
                    relpath = os.path.relpath(file_path, path)
                    z.write(file_path, relpath)

    def test_plain_zip(self):
        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            ori_files_dir = os.path.join(tmp_folder, "subfolder-1.2.3")
            file1 = os.path.join(ori_files_dir, "file1")
            file2 = os.path.join(ori_files_dir, "folder", "file2")
            file3 = os.path.join(ori_files_dir, "file3")

            save(file1, "")
            save(file2, "")
            save(file3, "")

        zip_file = os.path.join(tmp_folder, "myzip.zip")
        # Zip with a "folder_entry" in the zip (not only for the files)
        self._zipdir(tmp_folder, zip_file, folder_entry="subfolder-1.2.3")

        # ZIP unzipped regularly
        extract_folder = temp_folder()
        output = RedirectedTestOutput()
        with redirect_output(output):
            unzip(ConanFileMock(), zip_file, destination=extract_folder, strip_root=False)
        assert "ERROR: Error extract" not in output
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3"))
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3", "file1"))
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3", "folder",
                                           "file2"))
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3", "file3"))

        # Extract without the subfolder
        extract_folder = temp_folder()
        output = RedirectedTestOutput()
        with redirect_output(output):
            unzip(ConanFileMock(), zip_file, destination=extract_folder, strip_root=True)
        assert "ERROR: Error extract" not in output
        assert not os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3"))
        assert os.path.exists(os.path.join(extract_folder, "file1"))
        assert os.path.exists(os.path.join(extract_folder, "folder", "file2"))
        assert os.path.exists(os.path.join(extract_folder, "file3"))

    def test_invalid_flat(self):
        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            # Not a single dir containing everything
            file1 = os.path.join(tmp_folder, "subfolder-1.2.3", "folder2", "file1")
            file2 = os.path.join(tmp_folder, "other-1.2.3", "folder", "file2")

            save(file1, "")
            save(file2, "")

        zip_folder = temp_folder()
        zip_file = os.path.join(zip_folder, "file.zip")
        self._zipdir(tmp_folder, zip_file)

        # Extract without the subfolder
        extract_folder = temp_folder()
        try:
            unzip(ConanFileMock(), zip_file, destination=extract_folder, strip_root=True)
            assert False, "Expected ConanException"
        except ConanException as e:
            assert "The zip file contains more than 1 folder in the root" in str(e)

    def test_invalid_flat_single_file(self):
        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            save("file1", "contentsfile1")

        zip_folder = temp_folder()
        zip_file = os.path.join(zip_folder, "file.zip")
        self._zipdir(tmp_folder, zip_file)

        # Extract without the subfolder
        extract_folder = temp_folder()
        try:
            unzip(ConanFileMock(), zip_file, destination=extract_folder, strip_root=True)
            assert False, "Expected ConanException"
        except ConanException as e:
            assert "The zip file contains a file in the root" in str(e)


class TestTarExtractPlain:

    def _compress_folder(self, folder, tgz_path, folder_entry=None):
        # Create a tar.gz file with the files in the folder and an additional TarInfo entry
        # for the folder_entry (the gather files doesn't return empty dirs)
        with open(tgz_path, "wb") as tgz_handle:
            tgz = gzopen_without_timestamps("name", fileobj=tgz_handle)
            if folder_entry:
                # Create an empty folder in the tgz file
                t = tarfile.TarInfo(folder_entry)
                t.mode = 488
                t.type = tarfile.DIRTYPE
                tgz.addfile(t)
            files, _ = gather_files(folder)
            for filename, abs_path in files.items():
                info = tarfile.TarInfo(name=filename)
                with open(os.path.join(folder, filename), 'rb') as file_handler:
                    tgz.addfile(tarinfo=info, fileobj=file_handler)
            tgz.close()

    def test_linkame_striproot_folder(self):
        tmp_folder = temp_folder()
        other_tmp_folder = temp_folder()
        save(os.path.join(other_tmp_folder, "foo.txt"), "")
        tgz_path = os.path.join(tmp_folder, "foo.tgz")

        with open(tgz_path, "wb") as tgz_handle:
            tgz = gzopen_without_timestamps("name", fileobj=tgz_handle)

            # Regular file
            info = tarfile.TarInfo(name="common/foo.txt")
            info.name = "common/subfolder/foo.txt"
            info.path = "common/subfolder/foo.txt"
            with open(os.path.join(other_tmp_folder, "foo.txt"), 'rb') as file_handler:
                tgz.addfile(tarinfo=info, fileobj=file_handler)

            # A hardlink to the regular file
            info = tarfile.TarInfo(name="common/foo.txt")
            info.linkname = "common/subfolder/foo.txt"
            info.linkpath = "common/subfolder/foo.txt"
            info.name = "common/subfolder/bar/foo.txt"
            info.path = "common/subfolder/bar/foo.txt"
            info.type = b'1'  # This indicates a hardlink to the tgz file "common/subfolder/foo.txt"
            tgz.addfile(tarinfo=info, fileobj=None)
            tgz.close()

        assert not os.path.exists(os.path.join(tmp_folder, "subfolder", "foo.txt"))
        assert not os.path.exists(os.path.join(tmp_folder, "subfolder", "bar", "foo.txt"))
        untargz(tgz_path, destination=tmp_folder, strip_root=True)
        assert os.path.exists(os.path.join(tmp_folder, "subfolder", "foo.txt"))
        assert os.path.exists(os.path.join(tmp_folder, "subfolder", "bar", "foo.txt"))

        # Check develop2 public unzip
        rmdir(os.path.join(tmp_folder, "subfolder"))
        assert not os.path.exists(os.path.join(tmp_folder, "subfolder", "foo.txt"))
        assert not os.path.exists(os.path.join(tmp_folder, "subfolder", "bar", "foo.txt"))
        unzip(ConanFileMock(), tgz_path, destination=tmp_folder, strip_root=True)
        assert os.path.exists(os.path.join(tmp_folder, "subfolder", "foo.txt"))
        assert os.path.exists(os.path.join(tmp_folder, "subfolder", "bar", "foo.txt"))

    def test_plain_tgz(self):

        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            # Create a couple of files
            ori_files_dir = os.path.join(tmp_folder, "subfolder-1.2.3")
            file1 = os.path.join(ori_files_dir, "file1")
            file2 = os.path.join(ori_files_dir, "folder", "file2")
            file3 = os.path.join(ori_files_dir, "file3")

            save(file1, "")
            save(file2, "")
            save(file3, "")

        tgz_folder = temp_folder()
        tgz_file = os.path.join(tgz_folder, "file.tar.gz")
        self._compress_folder(tmp_folder, tgz_file, folder_entry="subfolder-1.2.3")

        # Tgz unzipped regularly
        extract_folder = temp_folder()
        untargz(tgz_file, destination=extract_folder, strip_root=False)
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3"))
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3", "file1"))
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3", "folder",
                                           "file2"))
        assert os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3", "file3"))

        # Extract without the subfolder
        extract_folder = temp_folder()
        untargz(tgz_file, destination=extract_folder, strip_root=True)
        assert not os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3"))
        assert os.path.exists(os.path.join(extract_folder, "file1"))
        assert os.path.exists(os.path.join(extract_folder, "folder", "file2"))
        assert os.path.exists(os.path.join(extract_folder, "file3"))

    def test_plain_tgz_common_base(self):

        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            # Create a couple of files
            ori_files_dir = os.path.join(tmp_folder, "subfolder-1.2.3")
            file1 = os.path.join(ori_files_dir, "folder", "file1")
            file2 = os.path.join(ori_files_dir, "folder", "file2")
            file3 = os.path.join(ori_files_dir, "folder", "file3")

            save(file1, "")
            save(file2, "")
            save(file3, "")

        tgz_folder = temp_folder()
        tgz_file = os.path.join(tgz_folder, "file.tar.gz")
        self._compress_folder(tmp_folder, tgz_file)

        # Tgz unzipped regularly
        extract_folder = temp_folder()
        untargz(tgz_file, destination=extract_folder, strip_root=True)
        assert not os.path.exists(os.path.join(extract_folder, "subfolder-1.2.3"))
        assert os.path.exists(os.path.join(extract_folder, "folder", "file1"))
        assert os.path.exists(os.path.join(extract_folder, "folder", "file2"))
        assert os.path.exists(os.path.join(extract_folder, "folder", "file3"))

    def test_invalid_flat(self):
        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            # Not a single dir containing everything
            file1 = os.path.join(tmp_folder, "subfolder-1.2.3", "folder2", "file1")
            file2 = os.path.join(tmp_folder, "other-1.2.3", "folder", "file2")

            save(file1, "")
            save(file2, "")

        tgz_folder = temp_folder()
        tgz_file = os.path.join(tgz_folder, "file.tar.gz")
        self._compress_folder(tmp_folder, tgz_file)

        extract_folder = temp_folder()
        try:
            untargz(tgz_file, destination=extract_folder, strip_root=True)
            assert False, "Expected ConanException"
        except ConanException as e:
            assert "The tgz file contains more than 1 folder in the root" in str(e)

    def test_invalid_flat_single_file(self):
        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            save("file1", "contentsfile1")

        zip_folder = temp_folder()
        tgz_file = os.path.join(zip_folder, "file.tar.gz")
        self._compress_folder(tmp_folder, tgz_file)

        # Extract without the subfolder
        extract_folder = temp_folder()
        try:
            unzip(ConanFileMock(), tgz_file, destination=extract_folder, strip_root=True)
            assert False, "Expected ConanException"
        except ConanException as e:
            assert "Can't untar a tgz containing files in the root with strip_root enabled" in str(e)

    def test_invalid_flat_multiple_file(self):
        tmp_folder = temp_folder()
        with chdir(tmp_folder):
            save("file1", "contentsfile1")
            save("file2", "contentsfile2")

        zip_folder = temp_folder()
        tgz_file = os.path.join(zip_folder, "file.tar.gz")
        self._compress_folder(tmp_folder, tgz_file)

        # Extract without the subfolder
        extract_folder = temp_folder()
        try:
            unzip(ConanFileMock(), tgz_file, destination=extract_folder, strip_root=True)
            assert False, "Expected ConanException"
        except ConanException as e:
            assert "Can't untar a tgz containing files in the root with strip_root enabled" in str(e)


def _compress_root_folder(folder, tgz_path, root_folder_name="root"):
    """
    Compress the whole folder adding withing the tar file also the folders. For instance:

    $ tar -tf '/tmp/folder/file.tar.gz'
    root/
    root/parent/
    root/parent/bin/
    root/parent/bin/file1
    root/parent/bin/file2
    """
    with open(tgz_path, "wb") as tgz_handle:
        tgz = gzopen_without_timestamps("name", fileobj=tgz_handle)
        files, _ = gather_files(folder)
        tgz.add(root_folder_name, arcname=os.path.basename(root_folder_name))
        tgz.close()


def test_decompressing_folders_with_different_modes():
    """
    When compressed folders come with restrictive permissions, untar() function should not
    raise a PermissionError.

    Issue related: https://github.com/conan-io/conan/issues/17987
    """
    tmp_folder = temp_folder()
    tgz_folder = temp_folder()
    tgz_file = os.path.join(tgz_folder, "file.tar.gz")
    with chdir(tmp_folder):
        save("root/parent/bin/file1", "contentsfile1")
        save("root/parent/bin/file2", "contentsfile2")
        os.chmod(os.path.join("root", "parent"), mode=0o555)
        os.chmod(os.path.join("root", "parent", "bin"), mode=0o700)
        _compress_root_folder(tmp_folder, tgz_file, root_folder_name="root")

    # Tgz unzipped regularly
    extract_folder = temp_folder()
    # Do not raise any PermissionError
    untargz(tgz_file, destination=extract_folder, strip_root=True)


@pytest.mark.skipif(sys.version_info < (3, 13, 4), reason="requires Python 3.13.4 or higher")
@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_decompressing_using_long_path_prefix():
    """
    If we use the "\\?\" prefix, untar() function should not
    raise an OSError: [WinError 123] error.

    Issue related: https://github.com/conan-io/conan/issues/18574
    """
    tmp_folder = temp_folder()
    tgz_folder = temp_folder()
    tgz_file = os.path.join(tgz_folder, "file.tar.gz")
    with chdir(tmp_folder):
        save("root/parent/bin/file1", "contentsfile1")
        save("root/parent/bin/file2", "contentsfile2")
        _compress_root_folder(tmp_folder, tgz_file, root_folder_name="root")

    # Tgz unzipped regularly
    extract_folder = f"\\\\?\\{temp_folder()}"
    # Do not raise any OSError: [WinError 123]
    untargz(tgz_file, destination=extract_folder, strip_root=True)
    assert "contentsfile1" in load(ConanFileMock(), os.path.join(extract_folder, "parent", "bin", "file1"))
    assert "contentsfile2" in load(ConanFileMock(), os.path.join(extract_folder, "parent", "bin", "file2"))
    untargz(tgz_file, destination=extract_folder, strip_root=False)
    assert "contentsfile1" in load(ConanFileMock(), os.path.join(extract_folder, "root", "parent", "bin", "file1"))
    assert "contentsfile2" in load(ConanFileMock(), os.path.join(extract_folder, "root", "parent", "bin", "file2"))



import os
import platform
from shutil import which

import pytest

from conan.internal.api.detect.detect_vs import vswhere
from conan.errors import ConanException
from conan.test.utils.env import environment_update


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Visual Studio")
@pytest.mark.tool("visual_studio")
class TestVisualStudioTools:

    def test_vswhere_not_found(self):
        """
        Locate vswhere in PATH or in ProgramFiles
        """
        # vswhere not found
        with environment_update({"ProgramFiles": None, "ProgramFiles(x86)": None, "PATH": ""}):
            with pytest.raises(ConanException) as e:
                vswhere()
            assert "Cannot locate vswhere" in str(e.value)

    def test_vswhere_path(self):
        # vswhere in ProgramFiles but not in PATH
        program_files = os.environ.get("ProgramFiles(x86)") or os.environ.get("ProgramFiles")
        vswhere_path = None
        if program_files:
            expected_path = os.path.join(program_files, "Microsoft Visual Studio", "Installer",
                                         "vswhere.exe")
            if os.path.isfile(expected_path):
                vswhere_path = expected_path
                with environment_update({"PATH": ""}):
                    assert vswhere()

        # vswhere in PATH but not in ProgramFiles
        env = {"ProgramFiles": None, "ProgramFiles(x86)": None}
        if not which("vswhere") and vswhere_path:
            env.update({"PATH": os.path.dirname(vswhere_path)})
        with environment_update(env):
            assert vswhere()



import os
import re
import textwrap
from collections import OrderedDict

import pytest

from conan.internal.paths import CONANFILE
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestServer


class TestSource:

    def test_local_flow_patch(self):
        # https://github.com/conan-io/conan/issues/2327
        conanfile = """from conan import ConanFile
from conan.tools.files import save, load
import os
class TestexportConan(ConanFile):
    name = "test"
    version = "0.1"

    exports = "mypython.py"
    exports_sources = "patch.patch"

    def source(self):
        save(self, "hello/hello.h", "my hello header!")
        patch = os.path.join(self.source_folder, "patch.patch")
        self.output.info("PATCH: %s" % load(self, patch))
        header = os.path.join(self.source_folder, "hello/hello.h")
        self.output.info("HEADER: %s" % load(self, header))
        python = os.path.join(self.recipe_folder, "mypython.py")
        self.output.info("PYTHON: %s" % load(self, python))
"""
        client = TestClient(light=True)
        client.save({"conanfile.py": conanfile,
                     "patch.patch": "mypatch",
                     "mypython.py": "mypython"})
        client.run("source .")
        assert "conanfile.py (test/0.1): PATCH: mypatch" in client.out
        assert "conanfile.py (test/0.1): HEADER: my hello header!" in client.out
        assert "conanfile.py (test/0.1): PYTHON: mypython" in client.out

        client.run("create . ")
        assert "test/0.1: PATCH: mypatch" in client.out
        assert "test/0.1: HEADER: my hello header!" in client.out
        assert "test/0.1: PYTHON: mypython" in client.out

    def test_apply_patch(self):
        # https://github.com/conan-io/conan/issues/2327
        # Test if a patch can be applied in source() both in create
        # and local flow
        client = TestClient(light=True)
        conanfile = """from conan import ConanFile
from conan.tools.files import load
import os
class Pkg(ConanFile):
    exports_sources = "*"
    def source(self):
        patch = os.path.join(self.source_folder, "mypatch")
        self.output.info("PATCH: %s" % load(self, patch))
"""
        client.save({"conanfile.py": conanfile,
                     "mypatch": "this is my patch"})
        client.run("source .")
        assert "PATCH: this is my patch" in client.out
        client.run("source .")
        assert "PATCH: this is my patch" in client.out
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "PATCH: this is my patch" in client.out

    def test_source_warning_os_build(self):
        # https://github.com/conan-io/conan/issues/2368
        conanfile = '''from conan import ConanFile
class ConanLib(ConanFile):
    pass
'''
        client = TestClient(light=True)
        client.save({CONANFILE: conanfile})
        client.run("source .")
        assert "This package defines both 'os' and 'os_build'" not in client.out

    def test_source_with_path_errors(self):
        client = TestClient(light=True)
        client.save({"conanfile.txt": "contents"}, clean_first=True)

        # Path with conanfile.txt
        client.run("source conanfile.txt", assert_error=True)
        assert "A conanfile.py is needed, %s is not acceptable" \
                      % os.path.join(client.current_folder, "conanfile.txt") in client.out

    def test_source_local_cwd(self):
        conanfile = '''
import os
from conan import ConanFile

class ConanLib(ConanFile):
    name = "hello"
    version = "0.1"

    def source(self):
        self.output.info("Running source!")
        self.output.info("cwd=>%s" % os.getcwd())
'''
        client = TestClient(light=True)
        client.save({CONANFILE: conanfile})

        client.run("install .")
        client.run("source .")
        assert "conanfile.py (hello/0.1): Calling source()" in client.out
        assert "conanfile.py (hello/0.1): cwd=>%s" % client.current_folder in client.out

    def test_local_source(self):
        conanfile = '''
from conan import ConanFile
from conan.tools.files import save

class ConanLib(ConanFile):

    def source(self):
        self.output.info("Running source!")
        err
        save(self, "file1.txt", "Hello World")
'''
        # First, failing source()
        client = TestClient(light=True)
        client.save({CONANFILE: conanfile})

        client.run("source .", assert_error=True)
        assert "conanfile.py: Running source!" in client.out
        assert "ERROR: conanfile.py: Error in source() method, line 9" in client.out

        # Fix the error and repeat
        client.save({CONANFILE: conanfile.replace("err", "")})
        client.run("source")
        assert "conanfile.py: Calling source() in" in client.out
        assert "conanfile.py: Running source!" in client.out
        assert "Hello World" == client.load("file1.txt")

    def test_local_source_layout_root(self):
        # Ensure that if a root folder is specified Conan source method
        # uses it. This is especially important when running with
        # no_copy_source where both the build()  and source() methods
        # need to reference source_folder (so the value must be the
        # same in both methods)
        conanfile = textwrap.dedent('''
            from conan import ConanFile
            from conan.tools.files import save, load

            class ConanLib(ConanFile):

                def layout(self):
                    self.folders.root = ".."
                    self.folders.source = "src"

                def source(self):
                    self.output.info(f"In the source() method the Source folder is: {self.source_folder}")
                    save(self, "source_file.c", "conent")

                def build(self):
                    self.output.info(f"In the build() method the Source folder is: {self.source_folder}")
                    load(self, f"{self.source_folder}/source_file.c")
            ''')
        client = TestClient(light=True)
        client.save({CONANFILE: conanfile})

        client.run("source .")
        # The build method will fail if source() and build() don't
        # both use the correct source_folder
        client.run("build .")
        assert "conanfile.py: Calling build()" in client.out  # doesn't fail

    def test_retrieve_exports_sources(self):
        # For Conan 2.0 if we install a package from a remote and we want to upload to other
        # remote we need to download the sources, as we consider revisions immutable, let's
        # iterate through the remotes to get the sources from the first match
        servers = OrderedDict()
        for index in range(2):
            servers[f"server{index}"] = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                                                   users={"user": "password"})

        client = TestClient(servers=servers, inputs=3*["user", "password"])
        client.save({"conanfile.py": GenConanfile().with_exports_sources("*"),
                     "sources.cpp": "sources"})
        client.run("create . --name=hello --version=0.1")
        rrev = client.exported_recipe_revision()
        client.run("upload hello/0.1 -r server0")
        # Ensure we uploaded it
        assert re.search(r"Uploading recipe 'hello/0.1#.*' \(.*\)", client.out)
        assert re.search(r"Uploading package 'hello/0.1#.*' \(.*\)", client.out)
        client.run("remove * -c")

        # install from server0 that has the sources, upload to server1 (does not have the package)
        # download the sources from server0
        client.run("install --requires=hello/0.1@ -r server0")
        client.run("upload hello/0.1 -r server1")
        assert "Sources downloaded from 'server0'" in client.out

        # install from server1 that has the sources, upload to server1
        # Will not download sources, revision already in server
        client.run("remove * -c")
        client.run("install --requires=hello/0.1@ -r server1")
        client.run("upload hello/0.1 -r server1")
        assert f"'hello/0.1#{rrev}' already in server, skipping upload" in client.out
        assert "Sources downloaded from 'server0'" not in client.out

        # install from server0 and build
        # download sources from server0
        client.run("remove * -c")
        client.run("install --requires=hello/0.1@ -r server0 --build='*'")
        assert "Sources downloaded from 'server0'" in client.out

    def test_source_method_called_once(self):
        """
        Test that the source() method will be executed just once, and the source code will
        be shared for all the package builds.
        """

        conanfile = textwrap.dedent('''
            import os
            from conan import ConanFile
            from conan.tools.files import save

            class ConanLib(ConanFile):

                def source(self):
                    save(self, os.path.join(self.source_folder, "main.cpp"), "void main() {}")
                    self.output.info("Running source!")
            ''')

        client = TestClient()
        client.save({CONANFILE: conanfile})

        client.run("create . --name=lib --version=1.0")
        assert "Running source!" in client.out

        client.run("create . --name=lib --version=1.0")
        assert "Running source!" not in client.out

        client.run("create . --name=lib --version=1.0 -s build_type=Debug")
        assert "Running source!" not in client.out

    def test_source_method_called_again_if_left_dirty(self):
        """
        If we fail in retreiving sources make sure the source() method will be called
        next time we create
        """

        conanfile = textwrap.dedent('''
            import os
            from conan import ConanFile

            class ConanLib(ConanFile):

                def source(self):
                    self.output.info("Running source!")
                    assert False
            ''')

        client = TestClient(light=True)
        client.save({CONANFILE: conanfile})

        client.run("create . --name=lib --version=1.0", assert_error=True)
        assert "Running source!" in client.out

        client.run("create . --name=lib --version=1.0", assert_error=True)
        assert "Running source!" in client.out
        assert "Source folder is corrupted, forcing removal" in client.out


class TestSourceWithoutDefaultProfile:
    # https://github.com/conan-io/conan/issues/12371
    @pytest.fixture()
    def client(self):
        c = TestClient()
        # The ``source()`` should still receive necessary configuration
        c.save_home({"global.conf": "tools.files.download:retry=MYCACHE"})
        # Make sure we don't have default profile
        os.remove(os.path.join(c.cache_folder, "profiles", "default"))
        return c

    def test_source_without_default_profile(self, client):
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                def source(self):
                    c = self.conf.get("tools.files.download:retry")
                    self.output.info("CACHE:{}!!".format(c))
                """)
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert "conanfile.py: Calling source()" in client.out
        assert "CACHE:MYCACHE!!" in client.out

    def test_source_with_layout(self, client):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import cmake_layout
            class Pkg(ConanFile):
                settings = "os", "compiler", "arch", "build_type"
                def layout(self):
                    cmake_layout(self)
                def source(self):
                    c = self.conf.get("tools.files.download:retry")
                    self.output.info("CACHE:{}!!".format(c))
                """)
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert "conanfile.py: Calling source()" in client.out
        assert "CACHE:MYCACHE!!" in client.out


def test_source_python_requires():
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile("pytool", "0.1")})
    c.run("export . ")
    c.run("upload * -r=default -c")
    c.run("remove * -c")

    c.save({"conanfile.py": GenConanfile().with_python_requires("pytool/0.1")}, clean_first=True)
    c.run("source . ")
    assert "pytool/0.1: Not found in local cache, looking in remotes" in c.out
    assert "pytool/0.1: Downloaded recipe" in c.out


@pytest.mark.parametrize("attribute", ["info", "settings", "options"])
def test_source_method_forbidden_attributes(attribute):
    conanfile = textwrap.dedent(f"""
    from conan import ConanFile
    class Package(ConanFile):
        def source(self):
            self.output.info(self.{attribute})
    """)
    client = TestClient(light=True)
    client.save({"conanfile.py": conanfile})

    client.run("source .", assert_error=True)
    assert f"'self.{attribute}' access in 'source()' method is forbidden" in client.out



import os

import pytest
import textwrap

from conan.test.assets.sources import gen_function_cpp
from conan.test.utils.tools import TestClient
from test.functional.toolchains.meson._base import check_binary


@pytest.mark.tool("ninja")
@pytest.mark.tool("meson")
@pytest.mark.tool("pkg_config")
class TestMesonPkgConfig:
    _conanfile_py = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.meson import Meson, MesonToolchain


    class App(ConanFile):
        settings = "os", "arch", "compiler", "build_type"
        generators = "PkgConfigDeps"
        requires = "hello/0.1"

        def layout(self):
            self.folders.build = "build"

        def generate(self):
            tc = MesonToolchain(self)
            tc.generate()

        def build(self):
            meson = Meson(self)
            meson.configure()
            meson.build()
    """)

    _meson_build = textwrap.dedent("""
    project('tutorial', 'cpp')
    hello = dependency('hello', version : '>=0.1')
    executable('demo', 'main.cpp', dependencies: hello)
    """)

    def test_reuse(self):
        t = TestClient()
        t.run("new cmake_lib -d name=hello -d version=0.1")
        t.run("create . -tf=\"\"")

        app = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])
        # Prepare the actual consumer package
        t.save({"conanfile.py": self._conanfile_py,
                "meson.build": self._meson_build,
                "main.cpp": app},
               clean_first=True)

        # Build in the cache
        t.run("build .")
        t.run_command(os.path.join("build", "demo"))

        assert "Hello World Release!" in t.out

        check_binary(t)



import os
import platform
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Windows", reason="Only windows")
def test_editable_msbuild():
    c = TestClient()
    c.run("new msbuild_lib -d name=dep -d version=0.1 -o dep")
    c.run("new cmake_exe -d name=pkg -d version=0.1 -d requires=dep/0.1 -o pkg")

    build_folder = ""
    output_folder = '--output-folder="{}"'.format(build_folder) if build_folder else ""
    build_folder = '--build-folder="{}"'.format(build_folder) if build_folder else ""

    def build_dep():
        c.run("install . {}".format(output_folder))
        c.run("build . {}".format(build_folder))
        c.run("install . -s build_type=Debug {}".format(output_folder))
        c.run("build . -s build_type=Debug {}".format(build_folder))

    with c.chdir("dep"):
        c.run("editable add . {}".format(output_folder))
        build_dep()

    def build_pkg(msg):
        c.run("build . ")
        folder = os.path.join("build", "Release")
        c.run_command(os.sep.join([".", folder, "pkg"]))
        assert "pkg/0.1: Hello World Release!" in c.out
        assert "{}: Hello World Release!".format(msg) in c.out
        c.run("build . -s build_type=Debug")
        folder = os.path.join("build", "Debug")
        c.run_command(os.sep.join([".", folder, "pkg"]))
        assert "pkg/0.1: Hello World Debug!" in c.out
        assert "{}: Hello World Debug!".format(msg) in c.out

    with c.chdir("pkg"):
        c.run("install . ")
        c.run("install . -s build_type=Debug")
        build_pkg("dep/0.1")

    # Do a source change in the editable!
    with c.chdir("dep"):
        dep_cpp = c.load("src/dep.cpp")
        dep_cpp = dep_cpp.replace("dep/", "SUPERDEP/")
        c.save({"src/dep.cpp": dep_cpp})
        build_dep()

    with c.chdir("pkg"):
        build_pkg("SUPERDEP/0.1")

    # Check that create is still possible
    c.run("editable remove dep")
    c.run("create dep")
    c.run("create pkg")
    # print(c.out)
    assert "pkg/0.1: Created package" in c.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
def test_editable_msbuilddeps():
    # https://github.com/conan-io/conan/issues/12521
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            name = "dep"
            version = "0.1"
            def layout(self):
                pass
            """)
    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_settings("build_type", "arch")
                                                          .with_requires("dep/0.1")
                                                          .with_generator("MSBuildDeps")})
    c.run("editable add dep")
    c.run("install pkg")
    # It doesn't crash anymore
    assert "dep/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709 - Editable" in c.out



# Test suite to check conan capabilities for cross compiling to web assembly and asmjs
import textwrap
import os
import platform
from shutil import rmtree
import pytest
import sys

from conan.test.utils.tools import TestClient

EMCC_MIN_PYTHON_VERSION = (3, 8)

base_emscripten_profile = textwrap.dedent(
    """
    [settings]
    build_type=Release
    compiler=emcc
    compiler.cppstd=17
    compiler.libcxx=libc++
    compiler.version=4.0.10
    os=Emscripten

    [conf]
    tools.build:exelinkflags=['-sALLOW_MEMORY_GROWTH=1']
    tools.build:sharedlinkflags=['-sALLOW_MEMORY_GROWTH=1']

    # Define the emcc executable paths
    tools.build:compiler_executables={'c':'emcc', 'cpp':'em++'}

    # Set Ninja as default generator as it is faster and will sove issues on Windows
    tools.cmake.cmaketoolchain:generator=Ninja
    # Verbosity to see emcc invocations
    tools.compilation:verbosity=verbose
    # Distinguish between architectures
    tools.cmake.cmake_layout:build_folder_vars=['settings.build_type', 'settings.arch']

    [buildenv]
    AR=emar
    NM=emnm
    RANLIB=emranlib
    STRIP=emstrip
"""
)

wasm32_profile = textwrap.dedent(
    """
    include(base_emscripten_profile)
    [settings]
    arch=wasm

    [conf]
    tools.build:exelinkflags+=['-sMAXIMUM_MEMORY=4GB', '-sINITIAL_MEMORY=64MB']
    tools.build:sharedlinkflags+=['-sMAXIMUM_MEMORY=4GB', '-sINITIAL_MEMORY=64MB']
    """
)

wasm_64_profile = textwrap.dedent(
    """
    include(base_emscripten_profile)
    [settings]
    arch=wasm64
    [conf]
    tools.build:exelinkflags+=['-sMAXIMUM_MEMORY=16GB', '-sINITIAL_MEMORY=16GB']
    tools.build:sharedlinkflags+=['-sMAXIMUM_MEMORY=16GB', '-sINITIAL_MEMORY=16GB']
    """
)


asmjs_profile = textwrap.dedent(
    """
    include(base_emscripten_profile)
    [settings]
    arch=asm.js

    [conf]
    tools.build:exelinkflags+=['-sMAXIMUM_MEMORY=2GB', '-sINITIAL_MEMORY=64MB']
    tools.build:sharedlinkflags+=['-sMAXIMUM_MEMORY=2GB', '-sINITIAL_MEMORY=64MB']
    """
)


@pytest.mark.tool("cmake")
@pytest.mark.tool("emcc")
@pytest.mark.tool("node")
@pytest.mark.skipif(sys.version_info < EMCC_MIN_PYTHON_VERSION, reason = "emcc requires Python 3.8 or higher")
@pytest.mark.skipif(platform.system() == "Windows", reason = "Emscripten not installed in Windows")
def test_cmake_emscripten():
    client = TestClient()

    client.run("new cmake_exe -d name=hello -d version=0.1")
    client.save({"wasm32": wasm32_profile, "asmjs": asmjs_profile, "base_emscripten_profile": base_emscripten_profile,})

    client.run("build . -pr:h=wasm32")
    assert "Conan toolchain: Defining libcxx as C++ flags: -stdlib=libc++" in client.out
    assert os.path.exists(os.path.join(client.current_folder, "build/release-wasm" , "hello.wasm"))

    # Run JavaScript generated code which uses .wasm file
    client.run_command("node ./build/release-wasm/hello")
    assert "Hello World Release!" in client.out

    client.run("build . -pr:h=asmjs")
    assert "WASM=0" in client.out
    # No wasm should have been generated for asm.js architecture
    assert not os.path.exists(os.path.join(client.current_folder, "build/release-asm.js" , "hello.wasm"))
    client.run_command("node ./build/release-asm.js/hello")
    assert "Hello World Release!" in client.out


@pytest.mark.tool("meson")
@pytest.mark.tool("emcc")
@pytest.mark.tool("node")
@pytest.mark.skipif(sys.version_info < EMCC_MIN_PYTHON_VERSION, reason = "emcc requires Python 3.8 or higher")
@pytest.mark.skipif(platform.system() == "Windows", reason = "Emscripten not installed in Windows")
def test_meson_emscripten():
    client = TestClient()
    client.run("new meson_exe -d name=hello -d version=0.1")

    client.save({"wasm32": wasm32_profile, "wasm64": wasm_64_profile, "asmjs": asmjs_profile, "base_emscripten_profile": base_emscripten_profile,})
    client.run("build . -pr:h=wasm64")
    assert "C++ compiler for the host machine: em++" in client.out
    assert "C++ linker for the host machine: em++ ld.wasm" in client.out
    assert "Host machine cpu family: wasm64" in client.out
    assert os.path.exists(os.path.join(client.current_folder, "build", "hello.wasm"))
    # wasm64 only supported since node v23 so only run in MacOS where it is available
    if platform.system() == "Darwin":
        client.run_command("node ./build/hello")
        assert "Hello World Release!" in client.out

    rmtree(os.path.join(client.current_folder, "build"))
    client.run("build . -pr:h=asmjs")
    assert "C++ compiler for the host machine: em++" in client.out
    assert "C++ linker for the host machine: em++ ld.wasm" in client.out
    assert "Host machine cpu family: asm.js" in client.out
    assert "WASM=0" in client.out

    assert not os.path.exists(os.path.join(client.current_folder, "build", "hello.wasm"))
    client.run_command("node ./build/hello")
    assert "Hello World Release!" in client.out


@pytest.mark.tool("autotools")
@pytest.mark.tool("emcc")
@pytest.mark.tool("node")
@pytest.mark.skipif(sys.version_info < EMCC_MIN_PYTHON_VERSION, reason = "emcc requires Python 3.8 or higher")
@pytest.mark.skipif(platform.system() == "Windows", reason = "Emscripten not installed in Windows")
def test_autotools_emscripten():
    client = TestClient(path_with_spaces=False)
    client.run("new autotools_exe -d name=hello -d version=0.1")
    client.save({"wasm32": wasm32_profile, "asmjs": asmjs_profile, "base_emscripten_profile": base_emscripten_profile,})
    client.run("build . -pr:h=wasm32")
    assert "checking for wasm32-local-emscripten-ranlib... emranlib" in client.out
    assert "checking for wasm32-local-emscripten-gcc... emcc" in client.out
    assert "checking for wasm32-local-emscripten-ar... emar" in client.out
    assert "checking the archiver (emar) interface... ar" in client.out
    assert "checking for wasm32-local-emscripten-strip... emstrip" in client.out

    assert os.path.exists(os.path.join(client.current_folder, "build-release", "src", "hello.wasm"))
    # Run JavaScript generated code which uses .wasm file
    client.run_command("node ./build-release/src/hello")
    assert "Hello World Release!" in client.out

    rmtree(os.path.join(client.current_folder, "build-release"))
    client.run("build . -pr:h=asmjs")
    assert "WASM=0" in client.out
    # No wasm should have been generated for asm.js architecture
    assert not os.path.exists(os.path.join(client.current_folder, "build-release", "hello.wasm"))
    client.run_command("node ./build-release/src/hello")
    assert "Hello World Release!" in client.out


@pytest.mark.tool("premake")
@pytest.mark.tool("emcc")
@pytest.mark.tool("node")
@pytest.mark.skipif(sys.version_info < EMCC_MIN_PYTHON_VERSION, reason = "emcc requires Python 3.8 or higher")
@pytest.mark.skipif(platform.system() != "Linux", reason = "Premake only installed in linux")
def test_premake_emscripten():
    client = TestClient()
    client.run("new premake_exe -d name=hello -d version=0.1")
    client.save({"wasm32": wasm32_profile, "asmjs": asmjs_profile, "base_emscripten_profile": base_emscripten_profile,})
    client.run("build . -pr:h=wasm32")
    assert "gmake --arch=wasm32" in client.out
    assert os.path.exists(os.path.join(client.current_folder, "build-release", "bin", "hello.wasm"))
    # Run JavaScript generated code which uses .wasm file
    client.run_command("node ./build-release/bin/hello")
    assert "Hello World Release!" in client.out
    rmtree(os.path.join(client.current_folder, "build-release"))

    client.run("build . -pr:h=asmjs")
    assert "WASM=0" in client.out
    # No wasm should have been generated for asm.js architecture
    assert not os.path.exists(os.path.join(client.current_folder, "build-release", "bin", "hello.wasm"))
    client.run_command("node ./build-release/bin/hello")
    assert "Hello World Release!" in client.out

# TODO: test_bazel_emscripten(): need WIP new bazel toolchain
# TODO: test_msbuild_emscripten(): give support to msbuild



import os

import pytest

from conan.api.model import RecipeReference
from conan.internal.paths import CONANINFO
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import load


@pytest.fixture()
def client():
    client = TestClient()
    conanfile = GenConanfile("mingw", "0.1").with_settings("compiler")
    build_msg = """
    def build(self):
        self.output.warning("COMPILER=> %s %s" % (self.name, str(self.settings.compiler)))
    """
    client.save({"conanfile.py": str(conanfile) + build_msg})
    client.run("export . --user=lasote --channel=testing")
    conanfile = GenConanfile("visual", "0.1").with_requires("mingw/0.1@lasote/testing")
    conanfile.with_settings("compiler")
    client.save({"conanfile.py": str(conanfile) + build_msg})
    client.run("export . --user=lasote --channel=testing")
    return client


def test_override(client):
    client.run("install --requires=visual/0.1@lasote/testing --build missing -s compiler=msvc "
               "-s compiler.version=191 -s compiler.runtime=dynamic "
               "-s mingw*:compiler='gcc' -s mingw*:compiler.libcxx='libstdc++' "
               "-s mingw*:compiler.version=4.8")

    assert "COMPILER=> mingw gcc" in client.out
    assert "COMPILER=> visual msvc" in client.out

    # CHECK CONANINFO FILE
    latest_rrev = client.cache.get_latest_recipe_revision(
        RecipeReference.loads("mingw/0.1@lasote/testing"))
    pkg_ids = client.cache.get_package_references(latest_rrev)
    latest_prev = client.cache.get_latest_package_revision(pkg_ids[0])
    package_path = client.cache.pkg_layout(latest_prev).package()
    conaninfo = load(os.path.join(package_path, CONANINFO))
    assert "compiler=gcc" in conaninfo

    # CHECK CONANINFO FILE
    package_path = client.created_layout().package()
    conaninfo = load(os.path.join(package_path, CONANINFO))
    assert "compiler=msvc" in conaninfo
    assert "compiler.version=191" in conaninfo


def test_non_existing_setting(client):
    client.run("install --requires=visual/0.1@lasote/testing --build missing -s compiler=msvc "
               "-s compiler.version=191 -s compiler.runtime=dynamic "
               "-s mingw/*:missingsetting='gcc' ", assert_error=True)
    assert "settings.missingsetting' doesn't exist" in client.out


def test_override_in_non_existing_recipe(client):
    client.run("install --requires=visual/0.1@lasote/testing --build missing -s compiler=msvc "
               "-s compiler.version=191 -s compiler.runtime=dynamic "
               "-s MISSINGID:compiler='gcc' ")

    assert "COMPILER=> mingw msvc" in client.out
    assert "COMPILER=> visual msvc" in client.out


def test_exclude_patterns_settings():

    client = TestClient()
    gen = GenConanfile().with_settings("build_type")
    client.save({"zlib/conanfile.py": gen})
    client.save({"openssl/conanfile.py": gen.with_require("zlib/1.0")})
    client.save({"consumer/conanfile.py": gen.with_require("openssl/1.0")})
    client.run("create zlib --name zlib --version 1.0")
    client.run("create openssl --name openssl --version 1.0")

    # We miss openssl and zlib debug packages
    client.run("install consumer -s build_type=Debug", assert_error=True)
    assert "ERROR: Missing prebuilt package for 'openssl/1.0', 'zlib/1.0'" in client.out

    # All except zlib are Release, the only missing is zlib debug
    client.run("install consumer -s build_type=Debug "
               "                 -s !zlib*:build_type=Release", assert_error=True)
    assert "ERROR: Missing prebuilt package for 'zlib/1.0'"

    # All the packages matches !potato* so all are Release
    client.run("install consumer -s build_type=Debug -s !potato*:build_type=Release")

    # All the packages except the consumer are Release, but we are creating consumer in Debug
    client.run("create consumer --name=consumer --version=1.0 "
               "-s=build_type=Debug -s=!&:build_type=Release")

    client.run("install --requires consumer/1.0 -s consumer/*:build_type=Debug")

    # Priority between package scoped settings
    client.run('remove consumer/*#* -c')
    client.run("install --reference consumer/1.0 -s build_type=Debug", assert_error=True)
    # Pre-check, there is no Debug package for any of them
    assert "ERROR: Missing prebuilt package for 'consumer/1.0', 'openssl/1.0', 'zlib/1.0'"
    # Pre-check there are Release packages
    client.run("create consumer --name=consumer --version=1.0 -s build_type=Release")

    # Try to install with this two scoped conditions, This is OK the right side has priority
    client.run("install --requires consumer/1.0 -s zlib/*:build_type=Debug -s *:build_type=Release")

    # Try to install with this two scoped conditions, This is ERROR the right side has priority
    client.run("install --requires consumer/1.0 -s *:build_type=Release -s zlib/*:build_type=Debug",
               assert_error=True)
    assert "ERROR: Missing prebuilt package for 'zlib/1.0'" in client.out

    # Try to install with this two scoped conditions, This is OK again, the right side has priority
    # The z* points to Release later, so zlib in Release
    client.run("install --requires consumer/1.0 -s *:build_type=Release "
               "-s zlib/*:build_type=Debug -s z*:build_type=Release")

    # Try to install with this two scoped conditions, This is OK again, the right side has priority
    # No package is potato, so all packages in Release
    client.run("install --requires consumer/1.0 -s !zlib:build_type=Debug "
               "-s !potato:build_type=Release")




from requests import Response

from conan.test.utils.tools import TestClient, TestServer, TestRequester

myconan1 = """
from conan import ConanFile
import platform

class HelloConan(ConanFile):
    name = "hello"
    version = "1.2.1"
"""


class TestDownload:

    def test_returns_on_failures(self):
        test_server = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")])
        servers = {"default": test_server}

        class BuggyRequester(TestRequester):
            def get(self, *args, **kwargs):
                resp = Response()
                resp.status_code = 404
                resp._content = b''
                return resp

        client2 = TestClient(servers=servers, requester_class=BuggyRequester)
        client2.run("remote add remotename url")
        client2.run("install --requires=foo/bar@ -r remotename", assert_error=True)
        assert "Package 'foo/bar' not resolved: Unable to find 'foo/bar' in remotes" in client2.out

        class BuggyRequester2(BuggyRequester):
            def get(self, *args, **kwargs):
                resp = Response()
                resp.status_code = 500
                resp._content = b'This server is under maintenance'
                return resp

        client2 = TestClient(servers=servers, requester_class=BuggyRequester2)
        client2.run("remote add remotename url")
        client2.run("install --requires=foo/bar@ -r remotename", assert_error=True)
        assert "ERROR: Package 'foo/bar' not resolved" in client2.out
        assert "This server is under maintenance" in client2.out
        assert "not found" not in client2.out



import textwrap

from conan.test.assets.sources import gen_function_cpp


def test_aggregator(transitive_libraries):
    c = transitive_libraries

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import CMake, cmake_layout
        class Pkg(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            requires = "engine/1.0"
            generators = "CMakeDeps", "CMakeToolchain"
            exports_sources = "*"
            def layout(self):
                cmake_layout(self)
            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
                self.run(os.path.join(self.cpp.build.bindir, "app"))
        """)

    cmakelists = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        project(app CXX)
        include(${CMAKE_BINARY_DIR}/generators/conandeps_legacy.cmake)
        add_executable(app main.cpp)
        target_link_libraries(app ${CONANDEPS_LEGACY})
        """)

    c.save({
        "conanfile.py": conanfile,
        "main.cpp": gen_function_cpp(name="main", includes=["engine"], calls=["engine"]),
        "CMakeLists.txt": cmakelists
    }, clean_first=True)
    c.run("build .")
    assert "matrix/1.0: Hello World Release!" in c.out
    assert "engine/1.0: Hello World Release!" in c.out



import textwrap
import time

import pytest

from conan.test.utils.tools import TestClient, GenConanfile

RUN_COUNT = 2


@pytest.mark.skip(reason="Compatibility performance test")
class TestCompatibilityPerformance:

    results = []

    @pytest.fixture(scope="module", autouse=True)
    def timings(self):
        results = []
        yield results

        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot()

        ax.set_title("Compatibility performance in real Artifactory server (last match, unoptimized)")

        #ax_tot = ax.twinx()

        # Plot a basic wireframe.

        grouping = {}
        x, y, y_tot_opt, y_tot_non = [], [], [], []
        stderr_opt, stderr_non = [], []
        for factors, run, with_optimization, duration in results:
            group = grouping.setdefault(factors, {"opt": [], "non": []})
            if with_optimization:
                group["opt"].append(duration)
            else:
                group["non"].append(duration)

        def ssd(values, mean):
            return sum((v - mean) ** 2 for v in values) / (len(values) - 1)

        for factors, durations in sorted(grouping.items()):
            x.append(factors)
            y.append(0)

            mean_opt = sum(durations["opt"]) / RUN_COUNT
            mean_non = sum(durations["non"]) / RUN_COUNT
            y_tot_opt.append(mean_opt)
            y_tot_non.append(mean_non)

            # ssd_opt = ssd(durations["opt"], mean_opt)
            # ssd_non = ssd(durations["non"], mean_non)
            # stderr_opt.append(ssd_opt / (RUN_COUNT ** 0.5))
            # stderr_non.append(ssd_non / (RUN_COUNT ** 0.5))

        max_diff_y = max(y)
        min_diff_y = min(min(y), 0)
        # ax_tot.set_ylim(ymin=min_diff_y-0.1*min_diff_y, ymax=max_diff_y+0.1*max_diff_y)

        max_tot_y = max(max(y_tot_non), max(y_tot_opt))
        ax.set_ylim(ymin=0, ymax=max_tot_y+0.1*max_tot_y)

        ax.grid(visible=True, which="major", axis="y", linestyle="--")
        # ax_tot.set_xticks(x)

        ax.set_xticks(x)
        ax.plot(x, y_tot_non, 'red', marker='x', label="Without optimization", linestyle=":")
        # ax.errorbar(x, y_tot_non, yerr=stderr_non)
        ax.plot(x, y_tot_opt, 'blue', marker='x', label="With optimization", linestyle=":")
        # ax.errorbar(x, y_tot_opt, yerr=stderr_opt)

        # ax_tot.plot(x, y, 'gray', marker='o', linestyle="--", label="Difference total")
        # ax_tot.set_ylabel("Total time diff (s)")

        ax.set_xlabel("Number of compatible configurations in server")
        ax.set_ylabel("Total time (s)")
        ax.legend(loc="upper left")

        # ax_tot.legend(loc="upper right")

        plt.show()

    @pytest.mark.parametrize("runs", range(RUN_COUNT))
    @pytest.mark.parametrize("with_optimization", [True, False])
    @pytest.mark.parametrize("factors", [pow(2, n) for n in range(1, 6)])
    def test_list_only_compatibility(self, timings, with_optimization, factors, runs):
        tc = TestClient(light=True)
        tc.run("remote add default https://conandev.jfrog.io/artifactory/api/conan/abril-tests")
        tc.run("remote login default abril")
        tc.run("remove * -c -r=default")
        cppstds = [7 + i*3 for i in range(factors+1)]
        compiler_settings = textwrap.dedent(f"""
                compiler:
                    foo:
                        version: [1]
                        cppstd: {cppstds}""")
        tc.run("version")
        tc.save_home({"settings_user.yml": compiler_settings})
        compat = tc.load_home("extensions/plugins/compatibility/compatibility.py")
        compat = compat.replace("cppstd_possible_values = supported_cppstd(conanfile)",
                                f"cppstd_possible_values = {cppstds}")
        tc.save_home({"extensions/plugins/compatibility/compatibility.py": compat})
        threshold = -1 if with_optimization else 1
        tc.save_home({"global.conf": f"user.graph:compatibility_optimization_threshold={factors + threshold}"})

        compiler_args = "-s compiler=foo -s compiler.version=1"
        pkg_name = str(time.time())
        tc.save({"conanfile.py": GenConanfile(pkg_name, "0.1").with_settings("compiler")})

        if with_optimization:
            for i in range(factors-1):
                tc.run(f"create . {compiler_args} -s compiler.cppstd={7 + i*3} -nr")
        tc.run(f"create . {compiler_args} -s compiler.cppstd={7 + (factors-1)*3} -nr")

        tc.run("upload * -r=default -c")
        tc.run("remove *:* -c")

        start_time = time.time()
        tc.run(f"install --requires={pkg_name}/0.1 {compiler_args} -s compiler.cppstd={7 + factors*3}")
        end_time = time.time()
        if with_optimization:
            assert f"Found {factors} compatible configurations in remotes" in tc.out
        else:
            assert f"compatible configurations in remotes" not in tc.out
        duration = end_time - start_time
        timings.append((factors, runs, with_optimization, duration))



import os
import pytest

from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import set_dirty, clean_dirty, set_dirty_context_manager, _DIRTY_FOLDER


class TestDirty:

    @pytest.fixture
    def setup_dirty(self):
        """ Create temporary folder to save dirty state
        """
        tempfolder = temp_folder()
        dirty_folder = tempfolder + _DIRTY_FOLDER
        return tempfolder, dirty_folder

    def test_set_dirty(self, setup_dirty):
        """ Dirty flag must be created by set_dirty
        """
        folder, dirty_folder = setup_dirty
        set_dirty(folder)
        assert os.path.exists(dirty_folder)

    def test_clean_dirty(self, setup_dirty):
        """ Dirty flag must be cleaned by clean_dirty
        """
        folder, dirty_folder = setup_dirty
        set_dirty(folder)
        assert os.path.exists(dirty_folder)
        clean_dirty(folder)
        assert not os.path.exists(dirty_folder)

    def test_set_dirty_context(self, setup_dirty):
        """ Dirty context must remove lock before exiting
        """
        folder, dirty_folder = setup_dirty
        with set_dirty_context_manager(folder):
            assert os.path.exists(dirty_folder)
        assert not os.path.exists(dirty_folder)

    def test_interrupted_dirty_context(self, setup_dirty):
        """ Broken context must preserve dirty state

            Raise an exception in middle of context. By default,
            dirty file is not removed.
        """
        folder, dirty_folder = setup_dirty
        try:
            with set_dirty_context_manager(folder):
                assert os.path.exists(dirty_folder)
                raise RuntimeError()
        except RuntimeError:
            pass
        assert os.path.exists(dirty_folder)



import os
import platform
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_local_static_generators_folder():
    """If we configure a generators folder in the layout, the generator files:
      - If belong to new generators: go to the specified folder: "my_generators"
      - If belong to old generators or txt: remains in the install folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type"))
    conan_file += """
    generators = "CMakeToolchain"
    def layout(self):
        self.folders.build = "build-{}".format(self.settings.build_type)
        self.folders.generators = "{}/generators".format(self.folders.build)
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -of=my_install")

    old_install_folder = os.path.join(client.current_folder, "my_install")
    cmake_toolchain_generator_path = os.path.join(old_install_folder, "conan_toolchain.cmake")
    assert not os.path.exists(cmake_toolchain_generator_path)

    build_folder = os.path.join(client.current_folder, "my_install", "build-Release")
    generators_folder = os.path.join(build_folder, "generators")
    cmake_toolchain_generator_path = os.path.join(generators_folder, "conan_toolchain.cmake")
    assert os.path.exists(cmake_toolchain_generator_path)


def test_local_dynamic_generators_folder():
    """If we configure a generators folder in the layout, the generator files:
      - If belong to new generators: go to the specified folder: "my_generators"
      - "txt" and old ones always to the install folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type").
                     with_import("from conan.tools.cmake import CMakeToolchain, CMake"))
    conan_file += """
    def generate(self):
        tc = CMakeToolchain(self)
        tc.generate()
    def layout(self):
        self.folders.build = "build-{}".format(self.settings.build_type)
        self.folders.generators = "{}/generators".format(self.folders.build)
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -of=my_install")

    old_install_folder = os.path.join(client.current_folder, "my_install")
    cmake_toolchain_generator_path = os.path.join(old_install_folder, "conan_toolchain.cmake")
    assert not os.path.exists(cmake_toolchain_generator_path)

    build_folder = os.path.join(client.current_folder, "my_install", "build-Release")
    generators_folder = os.path.join(build_folder, "generators")
    cmake_toolchain_generator_path = os.path.join(generators_folder, "conan_toolchain.cmake")
    assert os.path.exists(cmake_toolchain_generator_path)


def test_no_layout_generators_folder():
    """If we don't configure a generators folder in the layout, the generator files:
      - all go to the install_folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type").
                     with_import("from conan.tools.cmake import CMakeToolchain, CMake"))
    conan_file += """
    def generate(self):
        tc = CMakeToolchain(self)
        tc.generate()
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -of=my_install")

    old_install_folder = os.path.join(client.current_folder, "my_install")
    cmake_toolchain_generator_path = os.path.join(old_install_folder, "conan_toolchain.cmake")

    # In the install_folder
    assert os.path.exists(cmake_toolchain_generator_path)

    # But not in the base folder
    assert not os.path.exists(os.path.join(client.current_folder, "conan_toolchain.cmake"))


def test_local_build():
    """If we configure a build folder in the layout, the installed files in a "conan build ."
    go to the specified folder: "my_build"
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conan.tools.files import save"))
    conan_file += """
    def layout(self):
        self.folders.generators = "my_generators"
        self.folders.build = "my_build"
    def build(self):
        self.output.warning("Generators folder: {}".format(self.folders.generators_folder))
        save(self, "build_file.dll", "bar")
"""
    client.save({"conanfile.py": conan_file})
    client.run("build . --output-folder=my_install")
    dll = os.path.join(client.current_folder, "my_install", "my_build", "build_file.dll")
    assert os.path.exists(dll)


def test_local_build_change_base():
    """If we configure a build folder in the layout, the build files in a "conan build ."
    go to the specified folder: "my_build under the modified base one "common"
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conan.tools.files import save"))
    conan_file += """
    def layout(self):
        self.folders.build = "my_build"
    def build(self):
        save(self, "build_file.dll", "bar")
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . --output-folder=common")
    client.run("build . --output-folder=common")
    dll = os.path.join(client.current_folder, "common", "my_build", "build_file.dll")
    assert os.path.exists(dll)


def test_local_source():
    """The "conan source" is NOT affected by the --output-folder
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_import("from conan.tools.files import save"))
    conan_file += """
    def layout(self):
        self.folders.source = "my_source"
    def source(self):
        save(self, "downloaded.h", "bar")
    """
    client.save({"conanfile.py": conan_file})
    client.run("install . -of=my_install")
    client.run("source .")
    header = os.path.join(client.current_folder, "my_source", "downloaded.h")
    assert os.path.exists(header)


def test_export_pkg():
    """The export-pkg, calling the "package" method, follows the layout if `cache_package_layout` """
    client = TestClient()
    conan_file = str(GenConanfile("lib", "1.0")
                     .with_import("from conan.tools.files import copy, save"))
    conan_file += """
    no_copy_source = True
    def layout(self):
        self.folders.source = "my_source"
        self.folders.build = "my_build"
    def source(self):
        save(self, "downloaded.h", "bar")
    def build(self):
        save(self, "library.lib", "bar")
        save(self, "generated.h", "bar")
    def package(self):
        copy(self, "*.h", self.source_folder, self.package_folder)
        copy(self, "*.h", self.build_folder, self.package_folder)
        copy(self, "*.lib", self.build_folder, self.package_folder)
    """

    client.save({"conanfile.py": conan_file})
    client.run("source .")
    client.run("build .")
    client.run("export-pkg .")

    pf = client.created_layout().package()

    # Check the artifacts packaged
    assert os.path.exists(os.path.join(pf, "downloaded.h"))
    assert os.path.exists(os.path.join(pf, "generated.h"))
    assert os.path.exists(os.path.join(pf, "library.lib"))


def test_export_pkg_local():
    """The export-pkg, without calling "package" method, with local package, follows the layout"""
    client = TestClient()
    conan_file = str(GenConanfile("lib", "1.0")
                     .with_import("from conan.tools.files import copy, save"))
    conan_file += """
    no_copy_source = True
    def layout(self):
        self.folders.source = "my_source"
        self.folders.build = "my_build"
        self.folders.package = "my_package"
    def source(self):
        save(self, "downloaded.h", "bar")
    def build(self):
        save(self, "library.lib", "bar")
        save(self, "generated.h", "bar")
    def package(self):
        copy(self, "*.h", self.source_folder, self.package_folder)
        copy(self, "*.h", self.build_folder, self.package_folder)
        copy(self, "*.lib", self.build_folder, self.package_folder)
    """

    client.save({"conanfile.py": conan_file})
    client.run("source .")
    client.run("build .")

    # assert "WARN: Package folder: {}".format(pf) in client.out

    client.run("export-pkg .")
    pf = client.created_layout().package()

    # Check the artifacts packaged, THERE IS NO "my_package" in the cache
    assert "my_package" not in pf
    assert os.path.exists(os.path.join(pf, "downloaded.h"))
    assert os.path.exists(os.path.join(pf, "generated.h"))
    assert os.path.exists(os.path.join(pf, "library.lib"))

    # Doing a conan create: Same as export-pkg, there is "my_package" in the cache
    client.run("create . ")
    pf = client.created_layout().package()
    assert "my_package" not in pf
    assert os.path.exists(os.path.join(pf, "downloaded.h"))
    assert os.path.exists(os.path.join(pf, "generated.h"))
    assert os.path.exists(os.path.join(pf, "library.lib"))


def test_start_dir_failure():
    c = TestClient()
    c.run("new cmake_lib -d name=dep -d version=0.1")
    c.run("install .")
    if platform.system() != "Windows":
        gen_folder = os.path.join(c.current_folder, "build", "Release", "generators")
    else:
        gen_folder = os.path.join(c.current_folder, "build", "generators")
    expected_path = os.path.join(gen_folder, "conan_toolchain.cmake")
    assert os.path.exists(expected_path)
    os.unlink(expected_path)
    with c.chdir("build"):
        c.run("install ..")
    assert os.path.exists(expected_path)


def test_local_folders_without_layout():
    """ Test that the "conan install" can report the local source and build folder
    """
    # https://github.com/conan-io/conan/issues/10566
    client = TestClient()
    conan_file = textwrap.dedent("""
        from conan import ConanFile
        class Test(ConanFile):
            version = "1.2.3"
            name = "test"

            def layout(self):
                self.folders.source = "."
                self.folders.build = "build"

            def generate(self):
                self.output.info("generate sf: {}".format(self.source_folder))
                self.output.info("generate bf: {}".format(self.build_folder))
        """)

    client.save({"conanfile.py": conan_file})
    client.run("install .")
    assert "conanfile.py (test/1.2.3): generate sf: {}".format(client.current_folder) in client.out
    assert "conanfile.py (test/1.2.3): generate bf: {}".format(client.current_folder) in client.out
